#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include "linelet_format.h"
#include "image.h"
#include "memory.h"
#include "timer.h"
#include "util.h"
#include "log.h"

static enum linelet_format_format format;
static unsigned width, height;
static unsigned frames;

static char *output;

static void print_help(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "linelet gray [-s/--size widthxheight] [-F/--frames number of frames] [-f/--pix_fmt format] [-o/--output raw video output] [-h/--help This help]\n");
   linelet_log_printf(LINELET_LOG_INFO, "\tFormats: yuv444p yuv422p yuv444p10 yuv422p10\n");
}

static void parse_input(int argc, char *argv[])
{
   const struct option opts[] = {
      { "help", 0, NULL, 'h' },
      { "pix_fmt", 1, NULL, 'f' },
      { "frames", 1, NULL, 'F' },
      { "size", 1, NULL, 's' },
      { "output", 1, NULL, 'o' },
      { NULL },
   };

   const char *optstring = "hf:F:s:o:";

   for (;;)
   {
      int c = getopt_long(argc, argv, optstring, opts, NULL);
      if (c == -1)
         break;

      switch (c)
      {
         case 'h':
            print_help();
            exit(EXIT_SUCCESS);

         default:
         case '?':
            print_help();
            exit(EXIT_FAILURE);

         case 'f':
            format = linelet_format_parse(optarg, NULL);
            break;

         case 'F':
            frames = strtoul(optarg, NULL, 0);
            break;

         case 's':
            if (sscanf(optarg, "%ux%u", &width, &height) != 2)
            {
               print_help();
               exit(EXIT_FAILURE);
            }
            break;

         case 'o':
            free(output);
            output = strdup(optarg);
            break;
      }
   }

   if (optind < argc)
   {
      print_help();
      exit(EXIT_FAILURE);
   }

   if (!output)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing output file.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (!width || !height)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing --size.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (!frames)
   {
      linelet_log_printf(LINELET_LOG_INFO, "0 frames.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (format == LINELET_FMT_UNKNOWN)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing input pixel format --pix_fmt.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
}

static void fill_u16(uint16_t *buf, uint16_t v, unsigned samples)
{
   for (unsigned i = 0; i < samples; i++)
      buf[i] = v;
}

int gray_main(int argc, char *argv[])
{
   parse_input(argc, argv);

   FILE *out_file = strcmp(output, "-") ? fopen(output, "wb") : stdout;
   if (!out_file)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to open output file %s.\n", output);
      return EXIT_FAILURE;
   }

   unsigned subsample = linelet_format_to_subsample(format);
   unsigned bpp = linelet_format_to_bpp(format);

   size_t luma_size = width * height * bpp;
   size_t chroma_size = luma_size >> subsample;

   void *luma_buf = linelet_malloc(luma_size);
   if (!luma_buf)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to allocate buffers.\n");
      return EXIT_FAILURE;
   }

   if (bpp == 2)
      fill_u16(luma_buf, 512, luma_size >> 1);
   else
      memset(luma_buf, 0x80, luma_size);

   for (unsigned f = 0; f < frames; f++)
   {
      if (fwrite(luma_buf, 1, luma_size, out_file) != luma_size)
         break;
      if (fwrite(luma_buf, 1, chroma_size, out_file) != chroma_size)
         break;
      if (fwrite(luma_buf, 1, chroma_size, out_file) != chroma_size)
         break;
   }

   free(output);
   linelet_free(luma_buf);
   if (out_file && out_file != stdout)
      fclose(out_file);

   return 0;
}

