#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include "linelet_format.h"
#include "image.h"
#include "memory.h"
#include "timer.h"
#include "util.h"
#include "log.h"

static enum linelet_format_format format;
static unsigned width, height;
static unsigned slice_height;

static char *input;
static char *output;

static void print_help(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "linelet interleave [-S/--slice-height slice height] [-s/--size widthxheight] [-p/--pix_fmt format] [-o/--output raw YUV output] [-i/--input raw YUV input] [-h/--help This help]\n");
   linelet_log_printf(LINELET_LOG_INFO, "\tFormats: yuv444p yuv422p yuv444p10 yuv422p10\n");
}

static void parse_input(int argc, char *argv[])
{
   slice_height = 1;

   const struct option opts[] = {
      { "help", 0, NULL, 'h' },
      { "pix_fmt", 1, NULL, 'f' },
      { "size", 1, NULL, 's' },
      { "input", 1, NULL, 'i' },
      { "output", 1, NULL, 'o' },
      { "slice_height", 1, NULL, 'S' },
      { NULL },
   };

   const char *optstring = "hf:s:i:o:S:";

   for (;;)
   {
      int c = getopt_long(argc, argv, optstring, opts, NULL);
      if (c == -1)
         break;

      switch (c)
      {
         case 'h':
            print_help();
            exit(EXIT_SUCCESS);

         default:
         case '?':
            print_help();
            exit(EXIT_FAILURE);

         case 'f':
            format = linelet_format_parse(optarg, NULL);
            break;

         case 's':
            if (sscanf(optarg, "%ux%u", &width, &height) != 2)
            {
               print_help();
               exit(EXIT_FAILURE);
            }
            break;

         case 'S':
            slice_height = strtoul(optarg, NULL, 0);
            break;

         case 'i':
            free(input);
            input = strdup(optarg);
            break;

         case 'o':
            free(output);
            output = strdup(optarg);
            break;
      }
   }

   if (optind < argc)
   {
      print_help();
      exit(EXIT_FAILURE);
   }

   if (format == LINELET_FMT_UNKNOWN)
   {
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (!input)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing input file.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (!output)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing output file.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (!width || !height)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing --size.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (format == LINELET_FMT_UNKNOWN)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing input pixel format --pix_fmt.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
}

int interleave_main(int argc, char *argv[])
{
   parse_input(argc, argv);

   FILE *in_file = strcmp(input, "-") ? fopen(input, "rb") : stdin;
   if (!in_file)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to open input file %s.\n", input);
      return EXIT_FAILURE;
   }

   FILE *out_file = strcmp(output, "-") ? fopen(output, "wb") : stdout;
   if (!out_file)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to open output file %s.\n", output);
      return EXIT_FAILURE;
   }

   unsigned subsample = linelet_format_to_subsample(format);
   unsigned bpp = linelet_format_to_bpp(format);

   // Padding.
   size_t luma_size = width * ((height + slice_height - 1) & ~(slice_height - 1)) * bpp;
   size_t chroma_size = luma_size >> subsample;
   size_t luma_step = width * bpp;
   size_t chroma_step = luma_step >> subsample;

   uint8_t *luma_buf = linelet_malloc(luma_size);
   uint8_t *cb_buf = linelet_malloc(chroma_size);
   uint8_t *cr_buf = linelet_malloc(chroma_size);

   if (!luma_buf || !cb_buf || !cr_buf)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to allocate buffers.\n");
      return EXIT_FAILURE;
   }

   memset(luma_buf, 0, luma_size);
   memset(cb_buf, 0, chroma_size);
   memset(cr_buf, 0, chroma_size);

   luma_size = width * height * bpp;
   chroma_size = luma_size >> subsample;

   for (;;)
   {
      if (fread(luma_buf, 1, luma_size, in_file) != luma_size)
         break;
      if (fread(cb_buf, 1, chroma_size, in_file) != chroma_size)
         break;
      if (fread(cr_buf, 1, chroma_size, in_file) != chroma_size)
         break;

      for (unsigned y = 0; y < height; y += slice_height)
      {
         if (fwrite(luma_buf + y * luma_step, 1, luma_step * slice_height, out_file) != luma_step * slice_height)
            goto out;
         if (fwrite(cb_buf + y * chroma_step, 1, chroma_step * slice_height, out_file) != chroma_step * slice_height)
            goto out;
         if (fwrite(cr_buf + y * chroma_step, 1, chroma_step * slice_height, out_file) != chroma_step * slice_height)
            goto out;
      }
   }
out:

   linelet_free(luma_buf);
   linelet_free(cb_buf);
   linelet_free(cr_buf);
   if (in_file && in_file != stdin)
      fclose(in_file);
   if (out_file && out_file != stdout)
      fclose(out_file);

   return 0;
}

