#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <errno.h>
#include "timer.h"
#include "util.h"
#include "log.h"

#ifdef _WIN32
// Win32 headers.
#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#include <windows.h>
#include <ws2tcpip.h>
#define close(x) closesocket(x)
#else
// POSIX headers.
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/fcntl.h>
#endif

static unsigned port;
static char *target;
static int buffer;
static bool throttle;
static int ttl;
static bool loop;
static bool ipv6;
static bool nonblock;

static void print_help(void)
{
   linelet_log_printf(LINELET_LOG_INFO,
         "rtp_send <in file> [-h/--help this help] [--ipv6] [--ttl multicast TTL] [--loop multicast loop] [-b/--buffer UDP buffer size] [-p/--port UDP port] [-T/--throttle] [-t/--target receiving server] [-n/--nonblock]\n");
}

static void parse_input(int argc, char *argv[])
{
   const struct option opts[] = {
      { "help", 0, NULL, 'h' },
      { "port", 1, NULL, 'p' },
      { "buffer", 1, NULL, 'b' },
      { "target", 1, NULL, 't' },
      { "throttle", 0, NULL, 'T' },
      { "ttl", 1, NULL, 'l' },
      { "loop", 0, NULL, 'L' },
      { "ipv6", 0, NULL, '6' },
      { "nonblock", 0, NULL, 'n' },
      { NULL },
   };

   const char *optstring = "hp:t:b:Tn";

   for (;;)
   {
      int c = getopt_long(argc, argv, optstring, opts, NULL);
      if (c == -1)
         break;

      switch (c)
      {
         case 'h':
            print_help();
            exit(EXIT_SUCCESS);

         default:
         case '?':
            print_help();
            exit(EXIT_FAILURE);

         case 'n':
            nonblock = true;
            break;

         case 'l':
            ttl = strtol(optarg, NULL, 0);
            break;

         case '6':
            ipv6 = true;
            break;

         case 'L':
            loop = true;
            break;

         case 'p':
            port = strtoul(optarg, NULL, 0);
            break;

         case 'T':
            throttle = true;
            break;

         case 'b':
            buffer = strtol(optarg, NULL, 0);
            break;

         case 't':
            target = strdup(optarg);
            break;
      }
   }

   if (optind >= argc)
   {
      print_help();
      exit(EXIT_FAILURE);
   }

   if (!port)
   {
      print_help();
      exit(EXIT_FAILURE);
   }

   if (!target)
   {
      print_help();
      exit(EXIT_FAILURE);
   }
}

static int open_udp_socket(const char *host, unsigned port, struct addrinfo **res)
{
#ifdef _WIN32
   WSADATA wsa;
   if (WSAStartup(MAKEWORD(2, 2), &wsa) != 0)
   {
      WSACleanup();
      return -1;
   }
#endif

   int sock = -1;
   char port_str[32];
   snprintf(port_str, sizeof(port_str), "%u", port);
   *res = NULL;

   struct addrinfo hints = {
      .ai_family = ipv6 ? AF_INET6 : AF_INET,
      .ai_socktype = SOCK_DGRAM,
   };

   int ret = getaddrinfo(host, port_str, &hints, res);
   if (ret < 0)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "getaddrinfo(): %s\n", gai_strerror(ret));
      goto error;
   }

   sock = socket((*res)->ai_family, (*res)->ai_socktype, (*res)->ai_protocol);
   if (sock < 0)
   {
      perror("socket");
      goto error;
   }

   if (ttl)
   {
      char tmp = ttl;
      if (setsockopt(sock,
               ipv6 ? IPPROTO_IPV6 : IPPROTO_IP,
               ipv6 ? IPV6_MULTICAST_HOPS : IP_MULTICAST_TTL,
               (const void*)&tmp, sizeof(tmp)) < 0)
      {
         perror("setsockopt - ttl");
         goto error;
      }
   }

   if (loop)
   {
      char tmp = 1;
      if (setsockopt(sock, 
               ipv6 ? IPPROTO_IPV6 : IPPROTO_IP,
               ipv6 ? IPV6_MULTICAST_LOOP : IP_MULTICAST_LOOP,
               (const void*)&tmp, sizeof(tmp)) < 0)
      {
         perror("setsockopt - loop");
         goto error;
      }
   }

   if (buffer)
   {
      if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (const void*)&buffer, sizeof(int)) < 0)
         linelet_log_printf(LINELET_LOG_WARN, "Failed to set UDP buffer size.\n");
   }

   int size = 0;
   getsockopt(sock, SOL_SOCKET, SO_SNDBUF, (void*)&size, (socklen_t[]) { sizeof(int) });
   linelet_log_printf(LINELET_LOG_INFO, "UDP buffer size: %d\n", size);

   if (nonblock)
   {
#ifdef _WIN32
      u_long mode = 1;
      ioctlsocket(sock, FIONBIO, &mode);
#else
      if (fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK) < 0)
         linelet_log_printf(LINELET_LOG_WARN, "Failed to set non-blocking mode.\n");
#endif
   }

   return sock;

error:
   if (sock >= 0)
      close(sock);
   if (*res)
      freeaddrinfo(*res);
   return -1;
}

#define TS_PACKET_SIZE 188
#define TS_PACKETS 7
#define RTP_PAYLOAD_SIZE (TS_PACKET_SIZE * TS_PACKETS)
#define RTP_HEADER_SIZE 12
#define RTP_PACKET_SIZE (RTP_PAYLOAD_SIZE + RTP_HEADER_SIZE)
#define RTP_VERSION 2
#define RTP_PAYLOAD_TYPE 33 // MPEG2-TS
#define PCR_TIME_BASE 27000000
#define RTP_TIME_BASE 90000
#define PCR_TO_RTP_DIV (PCR_TIME_BASE / RTP_TIME_BASE)

// Parse adaptation field for PCR updates.
static bool parse_pcr(const uint8_t *payload, int64_t *pcr)
{
   unsigned adaptation = (payload[3] >> 4) & 3;
   if (adaptation <= 1)
      return false;

   payload += 4;
   unsigned len = *payload++;

   if (!len)
      return false;
   unsigned flags = *payload++;
   len--;
   if (!(flags & 0x10))
      return false;
   if (len < 6)
      return false;

   uint32_t v = linelet_read32_be(payload);
   int64_t hi = ((int64_t)v << 1) | (payload[4] >> 7);
   int64_t lo = ((payload[4] & 1) << 8) | payload[5];

   *pcr = hi * 300 + lo;
   return true;
}

static bool rtp_send(int sock, const struct addrinfo *addr,
      uint16_t seq, uint32_t timestamp, const uint8_t *payload, size_t size)
{
   uint8_t packet[RTP_PACKET_SIZE];

   packet[0] = RTP_VERSION << 6;
   packet[1] = RTP_PAYLOAD_TYPE;
   linelet_write16_be(packet + 2, seq);
   linelet_write32_be(packet + 4, timestamp);
   linelet_write32_be(packet + 8, 0);

   memcpy(packet + RTP_HEADER_SIZE, payload, size);

   ssize_t ret = sendto(sock, (const void*)packet, size + RTP_HEADER_SIZE, 0, addr->ai_addr, addr->ai_addrlen);
   if (ret < 0)
   {
#ifdef _WIN32
      if (WSAGetLastError() == WSAEWOULDBLOCK)
         linelet_log_printf(LINELET_LOG_WARN, "Got WSAEWOULDBLOCK error. Consider reducing bitrate or increasing buffer sizes.\n");
#else
      if (errno == EAGAIN)
         linelet_log_printf(LINELET_LOG_WARN, "Got EAGAIN error. Consider reducing bitrate or increasing buffer sizes.\n");
#endif
      else
      {
         perror("send");
         return false;
      }
   }

   return true;
}

int rtp_send_main(int argc, char *argv[])
{
   parse_input(argc, argv);

   struct addrinfo *res = NULL;
   int sock = open_udp_socket(target, port, &res);
   free(target);
   if (sock < 0)
      return EXIT_FAILURE;

   FILE *file = strcmp(argv[optind], "-") ? fopen(argv[optind], "rb") : stdin;
   if (!file)
   {
      close(sock);
      return EXIT_FAILURE;
   }

   // Keep track of where we found the last PCR. Used to extrapolate.
   int64_t last_pcr = -1;
   int64_t last_pcr_pos = -1;
   int64_t last_rate = 0; // Bytes / second. Used to extrapolate transmission times.

   int pcr_pid = -1; // Use the first PID that includes PCR as the reference clock.

   // Rate control.
   int64_t first_sent_time = -1;
   int64_t first_sent_pcr = -1;

   int64_t packets = 0;
   uint16_t seq = 0;

   for (;;)
   {
      uint8_t payload[RTP_PAYLOAD_SIZE];
      if (fread(payload, 1, sizeof(payload), file) != sizeof(payload))
         break;

      for (unsigned i = 0; i < TS_PACKETS; i++)
      {
         // TS sync byte. We assume the TS we're reading is pristine.
         if (payload[0] != 0x47)
            goto out;

         int pid = linelet_read16_be(payload + i * TS_PACKET_SIZE + 1) & 0x1fff;
         int64_t new_pcr = -1;
         if ((pcr_pid < 0 || pcr_pid == pid) && parse_pcr(payload + i * TS_PACKET_SIZE, &new_pcr))
         {
            // Use this reference for CBR timing.
            if (pcr_pid < 0)
            {
               first_sent_time = linelet_timer_get_nsec();
               first_sent_pcr = new_pcr;
            }

            pcr_pid = pid;
            int64_t pcr_pos = (packets + i) * TS_PACKET_SIZE;
            if (last_pcr_pos >= 0)
               last_rate = PCR_TIME_BASE * (pcr_pos - last_pcr_pos) / (new_pcr - last_pcr);

            last_pcr = new_pcr;
            last_pcr_pos = pcr_pos;
         }
      }

      // Compute RTP timestamp.
      int64_t transmission_pcr = 0;
      if (last_rate)
      {
         transmission_pcr = last_pcr +
            (PCR_TIME_BASE * (packets * TS_PACKET_SIZE - last_pcr_pos)) / last_rate;
      }

      //linelet_log_printf(LINELET_LOG_INFO, "Transmission PCR: %lld\n", (long long)transmission_pcr);

      uint32_t timestamp = transmission_pcr / PCR_TO_RTP_DIV;

      if (!rtp_send(sock, res, seq++, timestamp, payload, sizeof(payload)))
         break;

      // Check if we should throttle (sleep) to approximate CBR.
      if (last_rate && throttle)
      {
         int64_t offset_ms = (transmission_pcr - first_sent_pcr) / (PCR_TIME_BASE / 1000); // millisecond resolution is fine.
         int64_t target_time = first_sent_time + 1000000 * offset_ms;

         int64_t current = linelet_timer_get_nsec();

         // We're going too slow (100ms lag). Try to keep up.
         if (current - 100000000 > target_time)
         {
            // Jitter up 20ms.
            linelet_log_printf(LINELET_LOG_WARN,
                  "Lagging %"PRId64" ms behind target. Pressing forward to meet target.\n", (current - target_time) / 1000000);
            first_sent_time += 20000000;
            target_time += 20000000;
         }

         // Only sleep if we're at least 1ms ahead.
         if (current + 1000000 < target_time)
            linelet_timer_sleep_until(target_time);
      }

      packets += TS_PACKETS;
   }
out:

   if (file && file != stdin)
      fclose(file);
   close(sock);
   if (res)
      freeaddrinfo(res);
   return 0;
}

