#include "image.h"
#include "memory.h"
#include "linelet_format.h"
#include "log.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>
#include <string.h>
#include <stdbool.h>

static unsigned width, height;
static enum linelet_format_format format;
static bool total;

static void print_help(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "linelet ssim_video <raw YUV a> <raw YUV b> [-s/--size widthxheight] [-f/--pix_fmt format] [-h/--help This help] [-t/--total only report total]\n");
}

static void parse_input(int argc, char *argv[])
{
   const struct option opts[] = {
      { "help", 0, NULL, 'h' },
      { "pix_fmt", 1, NULL, 'f' },
      { "size", 1, NULL, 's' },
      { "total", 0, NULL, 't' },
      { NULL },
   };

   const char *optstring = "hf:s:t";

   for (;;)
   {
      int c = getopt_long(argc, argv, optstring, opts, NULL);
      if (c == -1)
         break;

      switch (c)
      {
         case 'h':
            print_help();
            exit(EXIT_SUCCESS);

         default:
         case '?':
            print_help();
            exit(EXIT_FAILURE);

         case 't':
            total = true;
            break;

         case 'f':
            format = linelet_format_parse(optarg, NULL);
            break;

         case 's':
            if (sscanf(optarg, "%ux%u", &width, &height) != 2)
            {
               print_help();
               exit(EXIT_FAILURE);
            }
            break;
      }
   }

   if (optind != argc - 2)
   {
      print_help();
      exit(EXIT_FAILURE);
   }

   if (format == LINELET_FMT_UNKNOWN)
   {
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (!width || !height)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing --size.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
   else if (format == LINELET_FMT_UNKNOWN)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Missing input pixel format --pix_fmt.\n");
      print_help();
      exit(EXIT_FAILURE);
   }
}

static double ssim_to_db(double ssim)
{
   return -10.0 * log10(1.0 - ssim);
}

int ssim_video_main(int argc, char *argv[])
{
   parse_input(argc, argv);

   FILE *a_file = strcmp(argv[optind + 0], "-") ? fopen(argv[optind + 0], "rb") : stdin;
   FILE *b_file = strcmp(argv[optind + 1], "-") ? fopen(argv[optind + 1], "rb") : stdin;
   if (!a_file || !b_file)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Failed to open files.\n");
      return EXIT_FAILURE;
   }

   unsigned subsample = linelet_format_to_subsample(format);
   unsigned bpp = linelet_format_to_bpp(format);
   unsigned bits = linelet_format_to_bits(format);

   size_t luma_size = width * height * bpp;
   size_t chroma_size = luma_size >> subsample;

   void *luma_buf_a = linelet_malloc(luma_size);
   void *cb_buf_a = linelet_malloc(chroma_size);
   void *cr_buf_a = linelet_malloc(chroma_size);
   void *luma_buf_b = linelet_malloc(luma_size);
   void *cb_buf_b = linelet_malloc(chroma_size);
   void *cr_buf_b = linelet_malloc(chroma_size);

   double ssim_luma_total = 0.0;
   double ssim_cb_total = 0.0;
   double ssim_cr_total = 0.0;
   unsigned ssim_frames = 0;

   for (;;)
   {
      if (fread(luma_buf_a, 1, luma_size, a_file) != luma_size)
         break;
      if (fread(cb_buf_a, 1, chroma_size, a_file) != chroma_size)
         break;
      if (fread(cr_buf_a, 1, chroma_size, a_file) != chroma_size)
         break;
      if (fread(luma_buf_b, 1, luma_size, b_file) != luma_size)
         break;
      if (fread(cb_buf_b, 1, chroma_size, b_file) != chroma_size)
         break;
      if (fread(cr_buf_b, 1, chroma_size, b_file) != chroma_size)
         break;

      double ssim_luma, ssim_cb, ssim_cr;
      if (bpp == 2)
      {
         ssim_luma = linelet_image_component_ssim_16(luma_buf_a, luma_buf_b, width, height, bits);
         ssim_cb = linelet_image_component_ssim_16(cb_buf_a, cb_buf_b, width >> subsample, height, bits);
         ssim_cr = linelet_image_component_ssim_16(cr_buf_a, cr_buf_b, width >> subsample, height, bits);
      }
      else
      {
         ssim_luma = linelet_image_component_ssim_8(luma_buf_a, luma_buf_b, width >> subsample, height);
         ssim_cb = linelet_image_component_ssim_8(cb_buf_a, cb_buf_b, width >> subsample, height);
         ssim_cr = linelet_image_component_ssim_8(cr_buf_a, cr_buf_b, width >> subsample, height);
      }

      ssim_luma_total += ssim_luma;
      ssim_cb_total += ssim_cb;
      ssim_cr_total += ssim_cr;
      ssim_frames++;

      if (!total)
      {
         printf("[SSIM]: Y: %11.6f, Cb: %11.6f, Cr: %11.6f\n",
               ssim_luma, ssim_cb, ssim_cr);
         printf("  [dB]: Y: %11.6f, Cb: %11.6f, Cr: %11.6f\n",
               ssim_to_db(ssim_luma), ssim_to_db(ssim_cb), ssim_to_db(ssim_cr));
      }
   }

   if (ssim_frames)
   {
      ssim_luma_total /= ssim_frames;
      ssim_cb_total /= ssim_frames;
      ssim_cr_total /= ssim_frames;

      printf("[SSIM Total]: Y: %11.6f, Cb: %11.6f, Cr: %11.6f\n",
            ssim_luma_total, ssim_cb_total, ssim_cr_total);
      printf("        [dB]: Y: %11.6f, Cb: %11.6f, Cr: %11.6f\n",
            ssim_to_db(ssim_luma_total), ssim_to_db(ssim_cb_total), ssim_to_db(ssim_cr_total));
   }

   if (a_file != stdin)
      fclose(a_file);
   if (b_file != stdin)
      fclose(b_file);

   return 0;
}
