#include "legall.h"
#include "image.h"
#include "quantizer.h"
#include "tests.h"
#include "timer.h"
#include "linelet.h"
#include "memory.h"
#include "simd.h"
#include "log.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#define min(a, b) ((a) < (b) ? (a) : (b))
#define max(a, b) ((a) > (b) ? (a) : (b))
#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))

static char *png_base;
static char *error_base;
static char *video_out;

static char **input_frames;
static unsigned num_input_frames;

static double rate = 0.5;
static unsigned slice_height = 1;
static unsigned threads = 1;
static unsigned precinct = 128;
static unsigned decomp = 5;
static unsigned iterations = 1;
static unsigned filter;

static enum linelet_encoder_rate_allocation_style style = LINELET_ENCODER_RATE_ALLOCATION_CBR;
static unsigned quant;

static void process_image(linelet_format_encoder_t *enc, linelet_format_decoder_t *dec,
      const char *png_base, const char *error_base,
      void *image_buffer, size_t target_size,
      struct linelet_rgb_image *out_img, struct linelet_rgb_image *in_img,
      double rate, unsigned slice_height)
{
   size_t encoded = 0;
   for (unsigned i = 0; i < iterations; i++)
   {
      TIMER_COUNTER_FORCE(encoder_process);
      TIMER_START_FORCE(encoder_process);
      encoded = linelet_format_encoder_process_frame(enc, (const void *[]) { in_img->rgb }, (const size_t[]) { in_img->stride }, image_buffer, target_size);
      TIMER_STOP_FORCE(encoder_process);
   }

   if (png_base || error_base)
   {
      for (unsigned i = 0; i < iterations; i++)
      {
         TIMER_COUNTER_FORCE(decoder_process);
         TIMER_START_FORCE(decoder_process);
         linelet_format_decoder_process_frame(dec, image_buffer, target_size,
               (void* const*)&out_img->rgb, &out_img->stride);
         TIMER_STOP_FORCE(decoder_process);
      }

      linelet_log_printf(LINELET_LOG_INFO, "\n==== Rate: %4.1f %% == Slice height: %2u == Target size: %u == Payload size: %u (%.2f %%) ====\n",
            100.0 * rate, slice_height, (unsigned)target_size, (unsigned)encoded,
            100.0 * encoded * rate / target_size);
   }

   if (png_base)
   {
      TIMER_COUNTER(save_image);
      TIMER_START(save_image);
      char out_path[PATH_MAX];
      snprintf(out_path, sizeof(out_path), "%s_%02u_%02u.png", png_base, slice_height, (unsigned)(rate * 100.0));
      linelet_log_printf(LINELET_LOG_INFO, "Saving image to %s.\n", out_path);
      out_img->bits = in_img->bits;
      if (!linelet_rgb_image_save(out_path, out_img))
         linelet_log_printf(LINELET_LOG_ERROR, "Failed to save image to %s.\n", out_path);
      TIMER_STOP(save_image);
   }

   if (error_base)
   {
      char out_path[PATH_MAX];
      snprintf(out_path, sizeof(out_path), "%s_%02u_%02u_error.png", error_base, slice_height, (unsigned)(rate * 100.0));
      linelet_log_printf(LINELET_LOG_INFO, "Saving error image to %s.\n", out_path);
      out_img->bits = in_img->bits;
      struct linelet_rgb_image error_image = {0};
      if (linelet_rgb_image_compute_error_image(&error_image, out_img, in_img))
      {
         if (!linelet_rgb_image_save(out_path, &error_image))
            linelet_log_printf(LINELET_LOG_ERROR, "Failed to save image to %s.\n", out_path);
         linelet_rgb_image_free(&error_image);
      }
   }

   if (png_base || error_base)
   {
      struct linelet_rgb_psnr psnr;
      if (linelet_rgb_image_psnr(&psnr, out_img, in_img, in_img->bits))
         linelet_log_printf(LINELET_LOG_INFO, "PSNR (dB) || R: %.1f, G: %.1f, B: %.1f\n", psnr.r, psnr.g, psnr.b);
   }

   linelet_log_printf(LINELET_LOG_INFO, "========================================================\n\n");
}

static void print_help(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "linelet [--iterations] [-S/--slice-height slice height] [-P/--precinct] [--filter] [-D/--decompositions] [-t/--threads num threads] [-r/--rate rate (0 to 1)] [-o/--LLV video output] [-p/--png PNG dump folder] [-h/--help This help] [-Q/--quant quantization factor (VBR)]\n");
}

static void parse_input(int argc, char *argv[])
{
   const struct option opts[] = {
      { "help", 0, NULL, 'h' },
      { "png", 1, NULL, 'p' },
      { "iterations", 1, NULL, 'I' },
      { "video", 1, NULL, 'o' },
      { "threads", 1, NULL, 't' },
      { "slice-height", 1, NULL, 'S' },
      { "rate", 1, NULL, 'r' },
      { "errorpng", 1, NULL, 'e' },
      { "quant", 1, NULL, 'Q' },
      { "precinct", 1, NULL, 'P' },
      { "decompositions", 1, NULL, 'D' },
      { "filter", 1, NULL, 'F' },
      { NULL },
   };

   const char *optstring = "hp:o:S:r:t:e:Q:P:D:";

   for (;;)
   {
      int c = getopt_long(argc, argv, optstring, opts, NULL);
      if (c == -1)
         break;

      switch (c)
      {
         case 'h':
            print_help();
            exit(EXIT_SUCCESS);

         default:
         case '?':
            print_help();
            exit(EXIT_FAILURE);

         case 'I':
            iterations = strtoul(optarg, NULL, 0);
            break;

         case 'F':
            filter = strtoul(optarg, NULL, 0);
            break;

         case 'p':
            png_base = strdup(optarg);
            break;

         case 'D':
            decomp = strtoul(optarg, NULL, 0);
            break;

         case 'P':
            precinct = strtoul(optarg, NULL, 0);
            break;

         case 'Q':
            quant = strtoul(optarg, NULL, 0);
            style = LINELET_ENCODER_RATE_ALLOCATION_MINIMUM_QUANT;
            break;

         case 'e':
            error_base = strdup(optarg);
            break;

         case 'r':
            rate = strtod(optarg, NULL);
            break;

         case 'S':
            slice_height = strtoul(optarg, NULL, 0); 
            break;

         case 't':
            threads = strtoul(optarg, NULL, 0);
            break;

         case 'o':
            video_out = strdup(optarg);
            break;
      }
   }

   int in_frames = argc - optind;
   if (!in_frames)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "No input frames!\n");
      print_help();
      exit(EXIT_FAILURE);
   }

   input_frames = linelet_calloc(in_frames, sizeof(char*));
   for (int i = 0; i < in_frames; i++)
      input_frames[i] = strdup(argv[optind + i]);
   num_input_frames = in_frames;
}

static void run_input_image(const char *path, FILE *video_file)
{
   struct linelet_rgb_image in_img = {0}, out_img = {0};
   if (!linelet_rgb_image_load(path, &in_img))
   {
      linelet_log_printf(LINELET_LOG_ERROR, "Failed to load image: %s.\n", path);
      exit(EXIT_FAILURE);
   }

   linelet_rgb_image_alloc(&out_img, in_img.width, in_img.height);

   linelet_format_encoder_t *enc = linelet_format_encoder_new(
         &(const struct linelet_format_encoder_options) {
            .format = LINELET_FMT_X2B10G10R10,
            .width = in_img.width,
            .height = in_img.height,
            .precinct_samples = precinct,
            .decomposition_levels = decomp,
            .slice_height = slice_height,
            .threads = threads,
            .rate_control = style,
            .q = quant,
            .filter = filter,
         });

   if (!enc)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "Failed to allocate encoder.\n");
      exit(EXIT_FAILURE);
   }

   linelet_format_decoder_t *dec = linelet_format_decoder_new(
         &(const struct linelet_format_decoder_options) {
            .format = LINELET_FMT_X2B10G10R10,
            .threads = threads,
         });

   if (!dec)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "Failed to allocate decoder.\n");
      exit(EXIT_FAILURE);
   }

   // Path handling in C is *so* much fun.
   char png_out_path[PATH_MAX];
   char error_out_path[PATH_MAX];
   if (png_base || error_base)
   {
      const char *path_base;
      const char *fslash = strrchr(path, '/');
      const char *bslash = strrchr(path, '\\');
      if (fslash && bslash)
         path_base = max(fslash, bslash) + 1;
      else if (fslash && !bslash)
         path_base = fslash + 1;
      else if (!fslash && bslash)
         path_base = bslash + 1;
      else
         path_base = path;

      snprintf(png_out_path, sizeof(png_out_path), "%s/%s", png_base, path_base);
      snprintf(error_out_path, sizeof(error_out_path), "%s/%s", error_base, path_base);
      char *last = strrchr(png_out_path, '.');
      if (last)
         *last = '\0';
      last = strrchr(error_out_path, '.');
      if (last)
         *last = '\0';
   }

   size_t target_size = in_img.width * in_img.height * rate * 3 * in_img.bits / 8;
   void *image_buffer = linelet_malloc(target_size);

   process_image(enc, dec,
         png_base ? png_out_path : NULL, error_base ? error_out_path : NULL,
         image_buffer, target_size,
         &out_img, &in_img, rate, slice_height);

   if (video_file)
   {
      if (fwrite(&target_size, sizeof(uint32_t), 1, video_file) != 1)
         linelet_log_printf(LINELET_LOG_ERROR, "Failed to write! (Disk full?)\n");

      if (fwrite(image_buffer, 1, target_size, video_file) != target_size)
         linelet_log_printf(LINELET_LOG_ERROR, "Failed to write! (Disk full?)\n");
   }

   linelet_rgb_image_free(&in_img);
   linelet_rgb_image_free(&out_img);

   linelet_format_encoder_free(enc);
   linelet_format_decoder_free(dec);
   linelet_free(image_buffer);
}

int test_main(int argc, char *argv[])
{
   tests_run_all();

   parse_input(argc, argv);

   FILE *video_file = NULL;

   // Ad-hoc format just for dumping rgb10a2 to disk so GL can read directly.
   if (video_out)
   {
      video_file = fopen(video_out, "wb");
      if (video_file)
      {
         char magic[] = "LLV1";
         if (fwrite(magic, 4, 1, video_file) != 1)
            linelet_log_printf(LINELET_LOG_ERROR, "Failed to write! (Disk full?)\n");
      }
      else
         linelet_log_printf(LINELET_LOG_ERROR, "Failed to open file.\n");
   }

   for (unsigned i = 0; i < num_input_frames; i++)
      run_input_image(input_frames[i], video_file);

   if (video_file)
      fclose(video_file);

   // Free out everything to make Valgrind happy.
   free(png_base);
   free(video_out);
   free(error_base);
   for (unsigned i = 0; i < num_input_frames; i++)
      free(input_frames[i]);
   linelet_free(input_frames);

   return 0;
}

