#ifndef LINELET_H__
#define LINELET_H__

#include <limits.h>
#include <stdint.h>
#include <stddef.h>
#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#else
#include <stdbool.h>
#endif

// Low-level raw bitstream encoder/decoder.

struct linelet_component
{
   unsigned decimation;
   unsigned decomposition_levels; // Number of wavelet decomposition levels. Number of wavelet bands == levels + 1.
};

#define LINELET_MAX_DECOMP_LEVELS 7
#define LINELET_MAX_BANDS (LINELET_MAX_DECOMP_LEVELS + 1)
#define LINELET_MAX_COMPONENTS 3
#define LINELET_MAX_LINES 16

struct linelet_wavelet_buffer
{
   int16_t *lp[LINELET_MAX_LINES][LINELET_MAX_DECOMP_LEVELS];
   int16_t *hp[LINELET_MAX_LINES][LINELET_MAX_DECOMP_LEVELS];
};

// Common
bool linelet_common_allocate_buffers(struct linelet_wavelet_buffer *buffer, unsigned samples, unsigned lines, const struct linelet_component *components, unsigned num_components);

void linelet_common_free_buffers(struct linelet_wavelet_buffer *buffers, unsigned num_components);

void linelet_common_set_component_levels(unsigned *levels, unsigned lines, unsigned ll_levels);

// Encoder
typedef struct linelet_encoder linelet_encoder_t;

enum linelet_encoder_rate_allocation_style
{
   // Every slice passed to the linelet_encoder will be compressed to maximum the buffer size.
   // If a large enough buffer is used, the encode is mathematically lossless.
   LINELET_ENCODER_RATE_ALLOCATION_CBR = 0,

   // Every slice will have a minimum degree of quantization applied (depending on q).
   // It will be further quantizied to fit inside the encoder buffer.
   // For "fixed quant" operation VBR-style, pass a large buffer to the encoder so it will always fit.
   // Q = 0, lowest quant, Q = 15 quantize everything to death.
   LINELET_ENCODER_RATE_ALLOCATION_MINIMUM_QUANT = 1,

   LINELET_ENCODER_RATE_ALLOCATION_DUMMY = INT_MAX
};

struct linelet_encoder_options
{
   unsigned precinct_samples; // Number of pixels per precinct.

   enum linelet_encoder_rate_allocation_style rate_control;
   unsigned q;
};

linelet_encoder_t *linelet_encoder_new(unsigned samples,
      unsigned lines,
      const struct linelet_component *components, unsigned num_components,
      const struct linelet_encoder_options *opts);

// Non-const component buffers because of inplace vertical lifting transform.
// If lines == 1, component_buffers will not be modified.
size_t linelet_encoder_process(linelet_encoder_t *ll,
      void *encoded_buffer, size_t encoded_buffer_size,
      int16_t * const *component_buffers, const size_t *component_strides);

void linelet_encoder_free(linelet_encoder_t *ll);


// Decoder
typedef struct linelet_decoder linelet_decoder_t;

struct linelet_decoder_options
{
   unsigned precinct_samples; // Number of pixels per precinct.
};

unsigned linelet_decoder_get_precinct_samples(const linelet_decoder_t *ll);

linelet_decoder_t *linelet_decoder_new(unsigned samples,
      unsigned lines,
      const struct linelet_component *components, unsigned num_components,
      const struct linelet_decoder_options *opts);

size_t linelet_decoder_process(linelet_decoder_t *ll,
      const void *encoded_buffer,
      int16_t * const *component_buffers, const size_t *component_strides);

void linelet_decoder_free(linelet_decoder_t *ll);

// Initializes linelet library. Does not have to be called by user.
void linelet_init(void);

#ifdef __cplusplus
}
#endif

#endif

