#ifndef TIMER_H__
#define TIMER_H__

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#else
#include <stdbool.h>
#endif

#if defined(__clang__)
#if SIZE_MAX > 0xffffffffu
#define TIMER_ATOMIC
#endif
#elif defined(__GNUC__)
#define GCC_VERSION (__GNUC__ * 100 + __GNUC_MINOR__)
#if (GCC_VERSION >= 407) && (SIZE_MAX > 0xffffffffu)
#define TIMER_ATOMIC
#endif
#endif

// Not thread safe, but avoids lots of nasty overhead. Only use when testing single-threaded perf.
#undef TIMER_ATOMIC

#define CONCAT_(a, b) a ## b
#define CONCAT(a, b) CONCAT_(a, b)

// Using relaxed mode here should be fine as we don't care about memory ordering until the very end.

#ifdef TIMER_ATOMIC
#define TIMER_INCREMENT_COUNT(x, cnt) __atomic_fetch_add(&CONCAT(x, counter).call_cnt, cnt, __ATOMIC_RELAXED)
#define TIMER_INCREMENT_TOTAL(x, delta) __atomic_fetch_add(&CONCAT(x, counter).total, delta, __ATOMIC_RELAXED)
#define TIMER_COUNTER_FORCE(x) static struct linelet_timer_counter CONCAT(x, counter) = { .desc = #x }; int64_t CONCAT(x, counter_start); \
   do { \
      if (!__atomic_test_and_set(&CONCAT(x, counter).registered, __ATOMIC_RELAXED)) \
         linelet_timer_register(&CONCAT(x, counter)); \
   } while(0)
#else
#define TIMER_COUNTER_FORCE(x) static struct linelet_timer_counter CONCAT(x, counter) = { .desc = #x }; int64_t CONCAT(x, counter_start); \
   do { \
      if (!CONCAT(x, counter).registered) linelet_timer_register(&CONCAT(x, counter)); \
   } while(0)
#define TIMER_INCREMENT_COUNT(x, cnt) \
   do { \
      CONCAT(x, counter).call_cnt += cnt; \
   } while(0)
#define TIMER_INCREMENT_TOTAL(x, delta) \
   do { \
      CONCAT(x, counter).total += delta; \
   } while(0)
#endif

struct linelet_timer_counter
{
   const char *desc;
   int64_t total;
   uint64_t call_cnt;
   bool registered;
};

#define TIMER_START_COUNT_FORCE(x, cnt) do { \
   TIMER_INCREMENT_COUNT(x, cnt); \
   CONCAT(x, counter_start) = linelet_timer_get_nsec(); \
} while(0)
#define TIMER_START_FORCE(x) TIMER_START_COUNT_FORCE(x, 1)
#define TIMER_STOP_FORCE(x) do { \
   int64_t delta = linelet_timer_get_nsec() - CONCAT(x, counter_start); \
   TIMER_INCREMENT_TOTAL(x, delta); \
} while(0)

#ifdef TIMER_ALL_ENABLE
#define TIMER_COUNTER(x) TIMER_COUNTER_FORCE(x)
#define TIMER_START_COUNT(x, cnt) TIMER_START_COUNT_FORCE(x, cnt)
#define TIMER_START(x) TIMER_START_FORCE(x)
#define TIMER_STOP(x) TIMER_STOP_FORCE(x)
#else
#define TIMER_COUNTER(x)
#define TIMER_START_COUNT(x, cnt)
#define TIMER_START(x)
#define TIMER_STOP(x)
#endif

int64_t linelet_timer_get_nsec(void);
int64_t linelet_timer_get_tick(void);
void linelet_timer_log(void);
void linelet_timer_register(struct linelet_timer_counter *cnt);

void linelet_timer_sleep_until(int64_t target);

#ifdef __cplusplus
}
#endif
#endif

