#include "convert.h"
#include "timer.h"
#include "simd.h"

void linelet_rct_forward_argb8888(int16_t * const *output, size_t stride,
      const uint32_t *input, size_t width, size_t height, size_t in_stride)
{
   TIMER_COUNTER(rct_forward_8bit);
   TIMER_START_COUNT(rct_forward_8bit, width * height * 3);

   linelet_rct_forward_argb8_yuv8_line_t simd = linelet_simd.rct_forward_argb8_yuv8_line;

   stride >>= 1;
   in_stride >>= 2;
   int16_t *l = output[0];
   int16_t *cb = output[1];
   int16_t *cr = output[2];

   for (size_t y = 0; y < height; y++, l += stride, cb += stride, cr += stride, input += in_stride)
   {
      unsigned x = 0;
      if (width >= 8)
      {
         x = width & ~7;
         simd(l, cb, cr, input, x);
      }

      for (; x < width; x++)
      {
         uint32_t col = input[x];
         int16_t r = (col >> 16) & 0xff;
         int16_t g = (col >>  8) & 0xff;
         int16_t b = (col >>  0) & 0xff;

         l[x] = ((r + (g << 1) + b) >> 2) - 128;
         cb[x] = b - g;
         cr[x] = r - g;
      }
   }

   linelet_simd.emms();
   TIMER_STOP(rct_forward_8bit);
}

void linelet_rct_forward_gbrp8(int16_t * const *output, size_t stride,
      const uint8_t * const *input, size_t width, size_t height, const size_t *in_stride)
{
   TIMER_COUNTER(rct_forward_8bit_planar);
   TIMER_START_COUNT(rct_forward_8bit_planar, width * height * 3);

   linelet_rct_forward_gbrp8_line_t simd = linelet_simd.rct_forward_gbrp8_line;

   stride >>= 1;
   int16_t *l = output[0];
   int16_t *cb = output[1];
   int16_t *cr = output[2];
   const uint8_t *G = input[0];
   const uint8_t *B = input[1];
   const uint8_t *R = input[2];

   for (size_t y = 0; y < height; y++,
         l += stride, cb += stride, cr += stride,
         G += in_stride[0], B += in_stride[1], R += in_stride[2])
   {
      unsigned x = 0;

      if (width >= 16)
      {
         x = width & ~15;
         simd(l, cb, cr, G, B, R, x);
      }

      for (; x < width; x++)
      {
         int16_t g = G[x];
         int16_t b = B[x];
         int16_t r = R[x];

         l[x] = ((r + (g << 1) + b) >> 2) - 128;
         cb[x] = b - g;
         cr[x] = r - g;
      }
   }

   linelet_simd.emms();
   TIMER_STOP(rct_forward_8bit_planar);
}

static inline int16_t clamp(int16_t v, int16_t min, int16_t max)
{
   if (v < min)
      return min;
   else if (v > max)
      return max;
   else
      return v;
}

void linelet_rct_inverse_argb8888(uint32_t *output, size_t out_stride, const int16_t * const *input, size_t width, size_t height, size_t stride)
{
   TIMER_COUNTER(rct_inverse_8bit);
   TIMER_START_COUNT(rct_inverse_8bit, width * height * 3);

   linelet_rct_inverse_yuv8_argb8_line_t simd = linelet_simd.rct_inverse_yuv8_argb8_line;

   stride >>= 1;
   out_stride >>= 2;
   const int16_t *l  = input[0];
   const int16_t *cb = input[1];
   const int16_t *cr = input[2];

   for (size_t y = 0; y < height; y++, l += stride, cb += stride, cr += stride, output += out_stride)
   {
      unsigned x = 0;
      if (width >= 8)
      {
         x = width & ~7;
         simd(output, l, cb, cr, x);
      }

      for (; x < width; x++)
      {
         int16_t l_samp  = l[x];
         int16_t cb_samp = cb[x];
         int16_t cr_samp = cr[x];

         int16_t g = l_samp - ((cb_samp + cr_samp) >> 2);
         int16_t r = cr_samp + g;
         int16_t b = cb_samp + g;

         g += 128;
         r += 128;
         b += 128;

         g = clamp(g, 0, 255);
         r = clamp(r, 0, 255);
         b = clamp(b, 0, 255);

         output[x] = (r << 16) | (g << 8) | (b << 0);
      }
   }

   linelet_simd.emms();
   TIMER_STOP(rct_inverse_8bit);
}

void linelet_rct_inverse_gbrp8(uint8_t * const *output, const size_t *out_stride, const int16_t * const *input, size_t width, size_t height, size_t stride)
{
   TIMER_COUNTER(rct_inverse_8bit);
   TIMER_START_COUNT(rct_inverse_8bit, width * height * 3);

   linelet_rct_inverse_gbrp8_line_t simd = linelet_simd.rct_inverse_gbrp8_line;

   stride >>= 1;
   const int16_t *l  = input[0];
   const int16_t *cb = input[1];
   const int16_t *cr = input[2];
   uint8_t *G = output[0];
   uint8_t *B = output[1];
   uint8_t *R = output[2];

   for (size_t y = 0; y < height; y++, l += stride, cb += stride, cr += stride,
         G += out_stride[0], B += out_stride[1], R += out_stride[2])
   {
      unsigned x = 0;
      if (width >= 16)
      {
         x = width & ~15;
         simd(G, B, R, l, cb, cr, x);
      }

      for (; x < width; x++)
      {
         int16_t l_samp  = l[x];
         int16_t cb_samp = cb[x];
         int16_t cr_samp = cr[x];

         int16_t g = l_samp - ((cb_samp + cr_samp) >> 2);
         int16_t r = cr_samp + g;
         int16_t b = cb_samp + g;

         g += 128;
         r += 128;
         b += 128;

         G[x] = clamp(g, 0, 255);
         B[x] = clamp(b, 0, 255);
         R[x] = clamp(r, 0, 255);
      }
   }

   linelet_simd.emms();
   TIMER_STOP(rct_inverse_8bit);
}

void linelet_rct_forward_a2bgr10_yuv10(int16_t * const *output, size_t stride,
      const uint32_t *input, size_t width, size_t height, size_t in_stride)
{
   TIMER_COUNTER(rct_forward_10bit);
   TIMER_START_COUNT(rct_forward_10bit, width * height * 3);

   linelet_rct_forward_a2bgr10_yuv10_line_t simd = linelet_simd.rct_forward_a2bgr10_yuv10_line;

   stride >>= 1;
   in_stride >>= 2;
   int16_t *l = output[0];
   int16_t *cb = output[1];
   int16_t *cr = output[2];

   for (size_t y = 0; y < height; y++, l += stride, cb += stride, cr += stride, input += in_stride)
   {
      const uint32_t *src = input;
      unsigned x = 0;
      if (width >= 8)
      {
         x = width & ~7;
         simd(l, cb, cr, src, x);
         src += x;
      }

      for (; x < width; x++)
      {
         uint32_t col = *src++;
         int16_t r = (col >>  0) & 0x3ff;
         int16_t g = (col >> 10) & 0x3ff;
         int16_t b = (col >> 20) & 0x3ff;

         l[x] = ((r + (g << 1) + b) >> 2) - 512;
         cb[x] = b - g;
         cr[x] = r - g;
      }
   }

   linelet_simd.emms();
   TIMER_STOP(rct_forward_10bit);
}

void linelet_rct_forward_gbrp10(int16_t * const *output, size_t stride,
      const uint16_t * const *input, size_t width, size_t height, const size_t *in_stride)
{
   TIMER_COUNTER(rct_forward_10bit_planar);
   TIMER_START_COUNT(rct_forward_10bit_planar, width * height * 3);

   linelet_rct_forward_gbrp10_line_t simd = linelet_simd.rct_forward_gbrp10_line;

   stride >>= 1;
   int16_t *l = output[0];
   int16_t *cb = output[1];
   int16_t *cr = output[2];
   const uint16_t *G = input[0];
   const uint16_t *B = input[1];
   const uint16_t *R = input[2];

   for (size_t y = 0; y < height; y++, l += stride, cb += stride, cr += stride,
         G += in_stride[0] >> 1, B += in_stride[1] >> 1, R += in_stride[2] >> 1)
   {
      unsigned x = 0;
      if (width >= 8)
      {
         x = width & ~7;
         simd(l, cb, cr, G, B, R, x);
      }

      for (; x < width; x++)
      {
         int16_t g = G[x];
         int16_t b = B[x];
         int16_t r = R[x];

         l[x] = ((r + (g << 1) + b) >> 2) - 512;
         cb[x] = b - g;
         cr[x] = r - g;
      }
   }

   linelet_simd.emms();
   TIMER_STOP(rct_forward_10bit_planar);
}

void linelet_rct_inverse_yuv10_a2bgr10(uint32_t *output, size_t out_stride, const int16_t * const *input, size_t width, size_t height, size_t stride)
{
   TIMER_COUNTER(rct_inverse_10bit);
   TIMER_START_COUNT(rct_inverse_10bit, width * height * 3);

   linelet_rct_inverse_yuv10_a2bgr10_line_t simd = linelet_simd.rct_inverse_yuv10_a2bgr10_line;

   stride >>= 1;
   out_stride >>= 2;
   const int16_t *l  = input[0];
   const int16_t *cb = input[1];
   const int16_t *cr = input[2];

   for (size_t y = 0; y < height; y++, l += stride, cb += stride, cr += stride, output += out_stride)
   {
      unsigned x = 0;
      if (width >= 8)
      {
         x = width & ~7;
         simd(output, l, cb, cr, x);
      }

      for (; x < width; x++)
      {
         int16_t l_samp  = l[x];
         int16_t cb_samp = cb[x];
         int16_t cr_samp = cr[x];

         int16_t g = l_samp - ((cb_samp + cr_samp) >> 2);
         int16_t r = cr_samp + g;
         int16_t b = cb_samp + g;

         g += 512;
         r += 512;
         b += 512;

         uint32_t r_out = clamp(r, 0, 1023);
         uint32_t g_out = clamp(g, 0, 1023);
         uint32_t b_out = clamp(b, 0, 1023);

         output[x] = (r_out << 0) | (g_out << 10) | (b_out << 20);
      }
   }

   linelet_simd.emms();
   TIMER_STOP(rct_inverse_10bit);
}

void linelet_rct_inverse_gbrp10(uint16_t * const *output, const size_t *out_stride, const int16_t * const *input,
      size_t width, size_t height, size_t stride)
{
   TIMER_COUNTER(rct_inverse_10bit_planar);
   TIMER_START_COUNT(rct_inverse_10bit_planar, width * height * 3);

   linelet_rct_inverse_gbrp10_line_t simd = linelet_simd.rct_inverse_gbrp10_line;

   stride >>= 1;
   const int16_t *l  = input[0];
   const int16_t *cb = input[1];
   const int16_t *cr = input[2];
   uint16_t *G = output[0];
   uint16_t *B = output[1];
   uint16_t *R = output[2];

   for (size_t y = 0; y < height; y++, l += stride, cb += stride, cr += stride,
         G += out_stride[0] >> 1, B += out_stride[1] >> 1, R += out_stride[2] >> 1)
   {
      unsigned x = 0;
      if (width >= 8)
      {
         x = width & ~7;
         simd(G, B, R, l, cb, cr, x);
      }

      for (; x < width; x++)
      {
         int16_t l_samp  = l[x];
         int16_t cb_samp = cb[x];
         int16_t cr_samp = cr[x];

         int16_t g = l_samp - ((cb_samp + cr_samp) >> 2);
         int16_t r = cr_samp + g;
         int16_t b = cb_samp + g;

         g += 512;
         r += 512;
         b += 512;

         G[x] = clamp(g, 0, 1023);
         B[x] = clamp(b, 0, 1023);
         R[x] = clamp(r, 0, 1023);
      }
   }

   linelet_simd.emms();
   TIMER_STOP(rct_inverse_10bit_planar);
}

void linelet_bt_yuv8_forward_dc_shift(int16_t * const *output, const size_t *stride,
      const uint8_t * const *input, size_t *width, size_t height, const size_t *in_stride)
{
   TIMER_COUNTER(bt_forward_8bit);
   TIMER_START_COUNT(bt_forward_8bit, width[0] * height * 3);

   linelet_bt_yuv8_forward_line_t simd = linelet_simd.bt_yuv8_forward_line;

   for (unsigned c = 0; c < 3; c++)
   {
      unsigned w = width[c];
      unsigned out_s = stride[c] >> 1;
      unsigned in_s = in_stride[c];
      const uint8_t *src = input[c];
      int16_t *dst = output[c];

      for (unsigned y = 0; y < height; y++, src += in_s, dst += out_s)
      {
         unsigned x = 0;
         if (w >= 16)
         {
            x = w & ~15;
            simd(dst, src, x);
         }

         for (; x < w; x++)
            dst[x] = src[x] - 128;
      }
   }

   linelet_simd.emms();
   TIMER_STOP(bt_forward_8bit);
}

void linelet_bt_yuv8_inverse_dc_shift(uint8_t * const *output, const size_t *stride,
      const int16_t * const *input, size_t *width, size_t height, const size_t *in_stride)
{
   TIMER_COUNTER(bt_inverse_8bit);
   TIMER_START_COUNT(bt_inverse_8bit, width[0] * height * 3);

   linelet_bt_yuv8_inverse_line_t simd = linelet_simd.bt_yuv8_inverse_line;

   for (unsigned c = 0; c < 3; c++)
   {
      unsigned w = width[c];
      unsigned out_s = stride[c];
      unsigned in_s = in_stride[c] >> 1;
      const int16_t *src = input[c];
      uint8_t *dst = output[c];

      for (unsigned y = 0; y < height; y++, src += in_s, dst += out_s)
      {
         unsigned x = 0;
         if (w >= 16)
         {
            x = w & ~15;
            simd(dst, src, x);
         }

         for (; x < w; x++)
            dst[x] = clamp(src[x] + 128, 0, 255); // FIXME: Determine if this should be TV range or PC range.
      }
   }

   linelet_simd.emms();
   TIMER_STOP(bt_inverse_8bit);
}

void linelet_bt_yuv10_forward_dc_shift(int16_t * const *output, const size_t *stride,
      const uint16_t * const *input, size_t *width, size_t height, const size_t *in_stride)
{
   TIMER_COUNTER(bt_forward_10bit);
   TIMER_START_COUNT(bt_forward_10bit, width[0] * height * 3);

   linelet_bt_yuv10_forward_line_t simd = linelet_simd.bt_yuv10_forward_line;

   for (unsigned c = 0; c < 3; c++)
   {
      unsigned w = width[c];
      unsigned out_s = stride[c] >> 1;
      unsigned in_s = in_stride[c] >> 1;
      const uint16_t *src = input[c];
      int16_t *dst = output[c];

      for (unsigned y = 0; y < height; y++, src += in_s, dst += out_s)
      {
         unsigned x = 0;
         if (w >= 16)
         {
            x = w & ~15;
            simd(dst, src, x);
         }

         for (; x < w; x++)
            dst[x] = src[x] - 512;
      }
   }

   linelet_simd.emms();
   TIMER_STOP(bt_forward_10bit);
}

void linelet_bt_yuv10_inverse_dc_shift(uint16_t * const *output, const size_t *stride,
      const int16_t * const *input, size_t *width, size_t height, const size_t *in_stride)
{
   TIMER_COUNTER(bt_inverse_10bit);
   TIMER_START_COUNT(bt_inverse_10bit, width[0] * height * 3);

   linelet_bt_yuv10_inverse_line_t simd = linelet_simd.bt_yuv10_inverse_line;

   for (unsigned c = 0; c < 3; c++)
   {
      unsigned w = width[c];
      unsigned out_s = stride[c] >> 1;
      unsigned in_s = in_stride[c] >> 1;
      const int16_t *src = input[c];
      uint16_t *dst = output[c];

      for (unsigned y = 0; y < height; y++, src += in_s, dst += out_s)
      {
         unsigned x = 0;
         if (w >= 16)
         {
            x = w & ~15;
            simd(dst, src, x);
         }

         for (; x < w; x++)
            dst[x] = clamp(src[x] + 512, 0, 0x3ff); // FIXME: Determine if this should be TV range or PC range.
      }
   }

   linelet_simd.emms();
   TIMER_STOP(bt_inverse_10bit);
}

void linelet_bt_yuv12_forward_dc_shift(int16_t * const *output, const size_t *stride,
      const uint16_t * const *input, size_t *width, size_t height, const size_t *in_stride)
{
   TIMER_COUNTER(bt_forward_12bit);
   TIMER_START_COUNT(bt_forward_12bit, width[0] * height * 3);

   linelet_bt_yuv12_forward_line_t simd = linelet_simd.bt_yuv12_forward_line;

   for (unsigned c = 0; c < 3; c++)
   {
      unsigned w = width[c];
      unsigned out_s = stride[c] >> 1;
      unsigned in_s = in_stride[c] >> 1;
      const uint16_t *src = input[c];
      int16_t *dst = output[c];

      for (unsigned y = 0; y < height; y++, src += in_s, dst += out_s)
      {
         unsigned x = 0;
         if (w >= 16)
         {
            x = w & ~15;
            simd(dst, src, x);
         }

         for (; x < w; x++)
            dst[x] = src[x] - 2048;
      }
   }

   linelet_simd.emms();
   TIMER_STOP(bt_forward_12bit);
}

void linelet_bt_yuv12_inverse_dc_shift(uint16_t * const *output, const size_t *stride,
      const int16_t * const *input, size_t *width, size_t height, const size_t *in_stride)
{
   TIMER_COUNTER(bt_inverse_12bit);
   TIMER_START_COUNT(bt_inverse_12bit, width[0] * height * 3);

   linelet_bt_yuv12_inverse_line_t simd = linelet_simd.bt_yuv12_inverse_line;

   for (unsigned c = 0; c < 3; c++)
   {
      unsigned w = width[c];
      unsigned out_s = stride[c] >> 1;
      unsigned in_s = in_stride[c] >> 1;
      const int16_t *src = input[c];
      uint16_t *dst = output[c];

      for (unsigned y = 0; y < height; y++, src += in_s, dst += out_s)
      {
         unsigned x = 0;
         if (w >= 16)
         {
            x = w & ~15;
            simd(dst, src, x);
         }

         for (; x < w; x++)
            dst[x] = clamp(src[x] + 2048, 0, 4095); // FIXME: Determine if this should be TV range or PC range.
      }
   }

   linelet_simd.emms();
   TIMER_STOP(bt_inverse_12bit);
}

