#include "linelet_format.h"
#include "linelet.h"
#include "image.h"
#include "convert.h"
#include "memory.h"
#include "haar.h"
#include "util.h"
#include "log.h"
#include "timer.h"
#include "checksum.h"
#include "prepostfilter.h"
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

struct thread_data
{
   linelet_decoder_t *dec;
   const uint8_t *payload;
   const uint8_t *payload_end;
   const uint8_t *scanline_meta;
   unsigned width;
   unsigned height;
   unsigned start_y;
   unsigned slice_height;
   unsigned real_height;

   uint8_t *dst[LINELET_MAX_COMPONENTS];
   size_t stride[LINELET_MAX_COMPONENTS];
   struct linelet_yuv_image img[2]; // For post filtering.

   pthread_t thread;
   pthread_mutex_t lock;
   pthread_cond_t cond;
   unsigned postfilter;
   bool ready;
   bool die;
   bool thread_active;

   void (*convert)(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines);
};

struct header_fmt1
{
   uint32_t width;
   uint32_t height;
   uint16_t color;
   uint16_t bits;
   uint16_t par_num;
   uint16_t par_den;
   uint32_t fps_num;
   uint32_t fps_den;
};

struct header_cod1
{
   uint16_t slice_height;
   uint16_t precinct_samples;
   unsigned num_components;
   unsigned postfilter;
   struct linelet_component components[LINELET_MAX_COMPONENTS];
};

struct header_ord1
{
   int64_t pts;
};

struct linelet_format_decoder
{
   struct header_fmt1 fmt1;
   struct header_cod1 cod1;
   struct header_ord1 ord1;

   struct linelet_format_decoder_options opts;
   enum linelet_format_format source_format;
   unsigned num_components;

   struct thread_data *threads;
   unsigned num_threads;

   const uint8_t *payload;
   const uint8_t *payload_end;
   const uint8_t *scanline_meta;
};

static void postfilter_slices(struct linelet_yuv_image *a,
      struct linelet_yuv_image *b, unsigned slice_height, unsigned postfilter)
{
   for (unsigned c = 0; c < LINELET_MAX_COMPONENTS; c++)
   {
      unsigned stride = a->strides[c] >> 1;

      switch (postfilter)
      {
         case 3:
            linelet_postfilter_8(
                  a->components[c] + (slice_height - 4) * stride,
                  a->components[c] + (slice_height - 3) * stride,
                  a->components[c] + (slice_height - 2) * stride,
                  a->components[c] + (slice_height - 1) * stride,
                  b->components[c] + 0 * stride,
                  b->components[c] + 1 * stride,
                  b->components[c] + 2 * stride,
                  b->components[c] + 3 * stride,
                  stride);
            break;

         case 2:
            linelet_postfilter_4(
                  a->components[c] + (slice_height - 2) * stride,
                  a->components[c] + (slice_height - 1) * stride,
                  b->components[c] + 0 * stride,
                  b->components[c] + 1 * stride,
                  stride);
            break;

         case 1:
            linelet_postfilter_2(
                  a->components[c] + (slice_height - 1) * stride,
                  b->components[c] + 0 * stride,
                  stride);
            break;
      }
   }
}

static void convert_x8r8g8b8(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   linelet_rct_inverse_argb8888((uint32_t*)thr->dst[0], thr->stride[0],
         (const int16_t**)img->components,
         thr->width, lines, img->luma.stride);
}

static void convert_x2b10g10r10(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   linelet_rct_inverse_yuv10_a2bgr10((uint32_t*)thr->dst[0], thr->stride[0],
         (const int16_t**)img->components,
         thr->width, lines, img->luma.stride);
}

static void convert_gbrp8(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   linelet_rct_inverse_gbrp8((uint8_t**)thr->dst, thr->stride,
         (const int16_t**)img->components,
         thr->width, lines, img->luma.stride);
}

static void convert_gbrp10(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   linelet_rct_inverse_gbrp10((uint16_t**)thr->dst, thr->stride,
         (const int16_t**)img->components,
         thr->width, lines, img->luma.stride);
}

static void convert_ycbcr_rct(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   for (unsigned c = 0; c < 3; c++)
      for (unsigned y = 0; y < lines; y++)
         memcpy(thr->dst[c] + y * thr->stride[c], img->components[c] + y * (img->strides[c] >> 1), img->strides[c]);
}

static void convert_ycbcr8_bt(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   size_t widths[] = {
      img->width >> img->luma.sub,
      img->width >> img->cb.sub,
      img->width >> img->cr.sub,
   };

   linelet_bt_yuv8_inverse_dc_shift((uint8_t**)thr->dst, thr->stride,
         (const int16_t**)img->components,
         widths, lines,
         img->strides);
}

static void convert_ycbcr10_bt(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   size_t widths[] = {
      img->width >> img->luma.sub,
      img->width >> img->cb.sub,
      img->width >> img->cr.sub,
   };

   linelet_bt_yuv10_inverse_dc_shift((uint16_t**)thr->dst, thr->stride,
         (const int16_t**)img->components,
         widths,
         lines, img->strides);
}

static void convert_ycbcr12_bt(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines)
{
   size_t widths[] = {
      img->width >> img->luma.sub,
      img->width >> img->cb.sub,
      img->width >> img->cr.sub,
   };

   linelet_bt_yuv12_inverse_dc_shift((uint16_t**)thr->dst, thr->stride,
         (const int16_t**)img->components,
         widths,
         lines, img->strides);
}

typedef void (*linelet_convert_func_t)(struct thread_data *thr, struct linelet_yuv_image *img, unsigned lines);
static linelet_convert_func_t format_to_convert_func(enum linelet_format_format fmt)
{
   switch (fmt)
   {
      case LINELET_FMT_X8R8G8B8:
         return convert_x8r8g8b8;
      case LINELET_FMT_X2B10G10R10:
         return convert_x2b10g10r10;
      case LINELET_FMT_GBRP8:
         return convert_gbrp8;
      case LINELET_FMT_GBRP10:
         return convert_gbrp10;

      case LINELET_FMT_YCBCR8_BT709_444:
      case LINELET_FMT_YCBCR8_BT709_422:
      case LINELET_FMT_YCBCR8_BT601_444:
      case LINELET_FMT_YCBCR8_BT601_422:
         return convert_ycbcr8_bt;

      case LINELET_FMT_YCBCR10_BT709_444:
      case LINELET_FMT_YCBCR10_BT709_422:
      case LINELET_FMT_YCBCR10_BT601_444:
      case LINELET_FMT_YCBCR10_BT601_422:
         return convert_ycbcr10_bt;

      case LINELET_FMT_YCBCR12_BT709_444:
      case LINELET_FMT_YCBCR12_BT709_422:
      case LINELET_FMT_YCBCR12_BT601_444:
      case LINELET_FMT_YCBCR12_BT601_422:
         return convert_ycbcr12_bt;

      case LINELET_FMT_YCBCR10_RCT_444:
      case LINELET_FMT_YCBCR10_RCT_422:
      case LINELET_FMT_YCBCR8_RCT_444:
      case LINELET_FMT_YCBCR8_RCT_422:
         return convert_ycbcr_rct;

      default:
         return NULL;
   }
}

static void decode_slice(struct thread_data *thr, struct linelet_yuv_image *img, uint32_t offset, uint32_t size)
{
   // Determine if slice is safe to decode.
   size_t decoded = 0;
   if (thr->payload + offset + size > thr->payload_end)
      linelet_log_printf(LINELET_LOG_ERROR, "[Decoder]: Payload out-of-bounds (corruption?)!\n");
   else
   {
      uint32_t checksum = linelet_read32_be(thr->payload + offset + size - sizeof(uint32_t));

      TIMER_COUNTER(decoder_checksum);
      TIMER_START(decoder_checksum);
      uint32_t computed = linelet_checksum(thr->payload + offset, size - sizeof(uint32_t));
      TIMER_STOP(decoder_checksum);

      if (checksum == computed)
      {
         decoded = linelet_decoder_process(thr->dec, thr->payload + offset,
               img->components, img->strides) + sizeof(uint32_t);
      }
      else
         linelet_log_printf(LINELET_LOG_ERROR, "[Decoder]: Checksum mismatch for slice.\n");
   }

   if (decoded != size)
      linelet_log_printf(LINELET_LOG_INFO, "[Decoder]: Decoder size differs from header (corruption?)!\n");
}

static void *thread_func(void *data)
{
   struct thread_data *thr = data;

   for (;;)
   {
      pthread_mutex_lock(&thr->lock);
      while (!thr->ready && !thr->die)
         pthread_cond_wait(&thr->cond, &thr->lock);
      pthread_mutex_unlock(&thr->lock);

      if (thr->die)
         break;

      unsigned img_index = 0;
      unsigned y = 0;
      unsigned overrun = 0;

      if (thr->postfilter)
      {
         // Need block from previous thread boundary.
         if (thr->start_y)
         {
            uint32_t offset = linelet_read32_be(thr->scanline_meta - 2 * sizeof(uint32_t));
            uint32_t size = linelet_read32_be(thr->scanline_meta - 1 * sizeof(uint32_t));
            decode_slice(thr, &thr->img[img_index ^ 1], offset, size);
         }

         // Decode first slice
         uint32_t offset = linelet_read32_be(thr->scanline_meta + 0);
         uint32_t size = linelet_read32_be(thr->scanline_meta + sizeof(uint32_t));
         decode_slice(thr, &thr->img[img_index], offset, size);

         // Have to postfilter against last slice over thread boundary.
         if (thr->start_y)
            postfilter_slices(&thr->img[img_index ^ 1], &thr->img[img_index], thr->slice_height, thr->postfilter);

         img_index ^= 1;
         thr->scanline_meta += 2 * sizeof(uint32_t);
         y = thr->slice_height;

         // We also need to decode next slice over the thread boundary to be able to output the last block.
         // Incrementing scanline_meta ensures that we get the overrun we want.
         // If we're not overrunning, compensate by lowering thr->height since we've already decoded one of the blocks.
         if (thr->start_y + thr->height < thr->real_height)
            overrun = thr->slice_height;
      }

      for (; y < thr->height + overrun; y += thr->slice_height)
      {
         uint32_t offset = linelet_read32_be(thr->scanline_meta + 0);
         uint32_t size = linelet_read32_be(thr->scanline_meta + sizeof(uint32_t));
         decode_slice(thr, &thr->img[img_index], offset, size);

         // We can postfilter against the block just decoded. Have to wait for next block as well to output this block however.
         if (thr->postfilter)
            postfilter_slices(&thr->img[img_index ^ 1], &thr->img[img_index], thr->slice_height, thr->postfilter);

         // We can now output the previous block if we used post filter.
         unsigned real_y = thr->start_y + y - (thr->postfilter ? thr->slice_height : 0);
         unsigned avail_y = LLMIN(thr->real_height - real_y, thr->slice_height);
         thr->convert(thr, &thr->img[img_index ^ !!thr->postfilter], avail_y);
         for (unsigned c = 0; c < LINELET_MAX_COMPONENTS; c++)
            thr->dst[c] += thr->slice_height * thr->stride[c];

         thr->scanline_meta += 2 * sizeof(uint32_t);
         if (thr->postfilter)
            img_index ^= 1;
      }

      // If we didn't overrun (last slice), we didn't get to output it, so flush it out now.
      if (thr->postfilter && !overrun)
      {
         unsigned real_y = thr->start_y + thr->height - thr->slice_height;
         unsigned avail_y = LLMIN(thr->real_height - real_y, thr->slice_height);
         thr->convert(thr, &thr->img[img_index ^ 1], avail_y);
      }

      pthread_mutex_lock(&thr->lock);
      thr->ready = false;
      pthread_cond_signal(&thr->cond);
      pthread_mutex_unlock(&thr->lock);
   }

   return NULL;
}

linelet_format_decoder_t *linelet_format_decoder_new(const struct linelet_format_decoder_options *opts)
{
   linelet_init();
   linelet_format_decoder_t *dec = linelet_calloc(1, sizeof(*dec));
   if (!dec)
      return NULL;

   unsigned threads = opts->threads ? opts->threads : 1;
   dec->threads = linelet_calloc(threads, sizeof(*dec->threads));
   if (!dec->threads)
      goto error;

   dec->num_components = linelet_format_to_components(opts->format);
   dec->num_threads = threads;
   for (unsigned i = 0; i < threads; i++)
   {
      struct thread_data *data = &dec->threads[i];

      pthread_mutex_init(&data->lock, NULL);
      pthread_cond_init(&data->cond, NULL);
      pthread_create(&data->thread, NULL, thread_func, data);
      data->thread_active = true;
   }

   dec->opts = *opts;
   return dec;

error:
   linelet_format_decoder_free(dec);
   return NULL;
}

void linelet_format_decoder_free(linelet_format_decoder_t *dec)
{
   if (!dec)
      return;

   for (unsigned i = 0; i < dec->num_threads; i++)
   {
      struct thread_data *data = &dec->threads[i];

      if (data->thread_active)
      {
         pthread_mutex_lock(&data->lock);
         data->ready = false;
         data->die = true;
         pthread_cond_signal(&data->cond);
         pthread_mutex_unlock(&data->lock);
         pthread_join(data->thread, NULL);
      }

      linelet_yuv_image_free(&data->img[0]);
      linelet_yuv_image_free(&data->img[1]);
      linelet_decoder_free(data->dec);
   }

   linelet_free(dec->threads);
   linelet_free(dec);
}

static const uint8_t *parse_fmt1(struct header_fmt1 *fmt1, const uint8_t *buffer)
{
   uint32_t len = linelet_read32_be(buffer);
   if (len != 24)
      return NULL;

   buffer += 4;
   fmt1->width   = linelet_read32_be(buffer + 0);
   fmt1->height  = linelet_read32_be(buffer + 4);
   fmt1->color   = linelet_read16_be(buffer + 8);
   fmt1->bits    = linelet_read16_be(buffer + 10);
   fmt1->par_num = linelet_read16_be(buffer + 12);
   fmt1->par_den = linelet_read16_be(buffer + 14);
   fmt1->fps_num = linelet_read32_be(buffer + 16);
   fmt1->fps_den = linelet_read32_be(buffer + 20);

   return buffer + 24;
}

static const uint8_t *parse_cod1(struct header_cod1 *cod1, const uint8_t *buffer)
{
   uint32_t len = linelet_read32_be(buffer);
   if (len != 3 + 3 * 2)
      return NULL;

   buffer += 4;
   cod1->precinct_samples = 1 << buffer[0];
   cod1->slice_height = 1 << buffer[1];
   cod1->num_components = buffer[2] & 0xf;
   cod1->postfilter = (buffer[2] >> 4) & 0xf;

   unsigned max_postfilter = linelet_log2i(cod1->slice_height);
   if (cod1->postfilter > max_postfilter)
   {
      linelet_log_printf(LINELET_LOG_ERROR, "Postfilter %u with slice_height == %u is invalid.\n", cod1->postfilter, cod1->slice_height);
      return NULL;
   }

   if (cod1->num_components != 3)
   {
      linelet_log_printf(LINELET_LOG_INFO, "Only 3 components supported.\n");
      return NULL;
   }

   buffer += 3;
   for (unsigned c = 0; c < cod1->num_components; c++, buffer += 2)
   {
      cod1->components[c].decimation = buffer[0];
      cod1->components[c].decomposition_levels = buffer[1];
   }

   return buffer;
}

static const uint8_t *parse_ord1(struct header_ord1 *ord1, const uint8_t *buffer)
{
   uint32_t len = linelet_read32_be(buffer);
   if (len != 8)
      return NULL;

   buffer += 4;
   ord1->pts = (int64_t)linelet_read64_be(buffer);
   return buffer + 8;
}

static const uint8_t *parse_off1(const uint8_t **scanline_meta, const struct header_fmt1 *fmt1, const struct header_cod1 *cod1, const uint8_t *buffer)
{
   uint32_t len = linelet_read32_be(buffer);
   if (len != ((fmt1->height + cod1->slice_height - 1) / cod1->slice_height) * sizeof(uint32_t) * 2)
      return NULL;

   buffer += 4;
   if (scanline_meta)
      *scanline_meta = buffer;
   return buffer + len;
}

static enum linelet_format_format bits_transform_decimation_to_format(unsigned bits, unsigned color, const struct linelet_component *components)
{
   bool subsample;
   if (components[0].decimation == 0 && components[1].decimation == 1 && components[2].decimation == 1)
      subsample = true;
   else if (components[0].decimation == 0 && components[1].decimation == 0 && components[2].decimation == 0)
      subsample = false;
   else
      return LINELET_FMT_UNKNOWN;
   
   if (bits != 8 && bits != 10 && bits != 12)
      return LINELET_FMT_UNKNOWN;

   switch (color)
   {
      case LINELET_YUV_RCT:
         if (bits == 10)
            return subsample ? LINELET_FMT_YCBCR10_RCT_422 : LINELET_FMT_YCBCR10_RCT_444;
         else
            return subsample ? LINELET_FMT_YCBCR8_RCT_422 : LINELET_FMT_YCBCR8_RCT_444;

      case LINELET_YUV_BT709:
         if (bits == 12)
            return subsample ? LINELET_FMT_YCBCR12_BT709_422 : LINELET_FMT_YCBCR12_BT709_444;
         else if (bits == 10)
            return subsample ? LINELET_FMT_YCBCR10_BT709_422 : LINELET_FMT_YCBCR10_BT709_444;
         else
            return subsample ? LINELET_FMT_YCBCR8_BT709_422 : LINELET_FMT_YCBCR8_BT709_444;

      case LINELET_YUV_BT601:
         if (bits == 12)
            return subsample ? LINELET_FMT_YCBCR12_BT601_422 : LINELET_FMT_YCBCR12_BT601_444;
         else if (bits == 10)
            return subsample ? LINELET_FMT_YCBCR10_BT601_422 : LINELET_FMT_YCBCR10_BT601_444;
         else
            return subsample ? LINELET_FMT_YCBCR8_BT601_422 : LINELET_FMT_YCBCR8_BT601_444;

      default:
         return LINELET_FMT_UNKNOWN;
   }
}

static bool parse_header(linelet_format_decoder_t *dec, const uint8_t *buffer, size_t encoded_buffer_size)
{
   const uint8_t *orig_buffer = buffer;
   if (memcmp(buffer, "LINELET1", 8))
      return false;
   buffer += 8;

   dec->ord1.pts = LINELET_FORMAT_PTS_INVALID;

   while ((size_t)(buffer + 8 - orig_buffer) < encoded_buffer_size)
   {
      if (!memcmp(buffer, "FMT1", 4))
      {
         buffer += 4;
         if (!(buffer = parse_fmt1(&dec->fmt1, buffer)))
            return false;
      }
      else if (!memcmp(buffer, "COD1", 4))
      {
         buffer += 4;
         if (!(buffer = parse_cod1(&dec->cod1, buffer)))
            return false;
      }
      else if (!memcmp(buffer, "ORD1", 4))
      {
         buffer += 4;
         if (!(buffer = parse_ord1(&dec->ord1, buffer)))
            return false;
      }
      else if (!memcmp(buffer, "OFF1", 4))
      {
         buffer += 4;
         if (!(buffer = parse_off1(&dec->scanline_meta, &dec->fmt1, &dec->cod1, buffer)))
            return false;
      }
      else if (!memcmp(buffer, "EOH1", 4))
      {
         uint32_t header_crc = linelet_checksum(orig_buffer, buffer - orig_buffer);
         uint32_t reference = linelet_read32_be(buffer + 4);

         if (reference != header_crc)
         {
            linelet_log_printf(LINELET_LOG_ERROR, "[Decoder]: Linelet header corrupt (CRC).\n");
            return false;
         }

         dec->payload = buffer + 8;
         dec->payload_end = orig_buffer + encoded_buffer_size;
         dec->source_format = bits_transform_decimation_to_format(dec->fmt1.bits, dec->fmt1.color, dec->cod1.components);
         return dec->source_format != LINELET_FMT_UNKNOWN;
      }
      else // Unknown field
      {
         linelet_log_printf(LINELET_LOG_INFO, "[Decoder]: Unknown TAG in header.\n");
         buffer += 4;
         buffer += linelet_read32_be(buffer);
      }
   }

   return false;
}

static bool header_to_frame_params(struct linelet_format_decoder_frame_params *params,
      const struct header_fmt1 *fmt1, const struct header_cod1 *cod1,
      const struct header_ord1 *ord1)
{
   params->width = fmt1->width;
   params->height = fmt1->height;
   params->sample_aspect_ratio.num = fmt1->par_num;
   params->sample_aspect_ratio.den = fmt1->par_den;
   params->frame_rate.num = fmt1->fps_num;
   params->frame_rate.den = fmt1->fps_den;
   params->format = bits_transform_decimation_to_format(fmt1->bits, fmt1->color, cod1->components);
   params->pts = ord1->pts;
   return params->format != LINELET_FMT_UNKNOWN;
}

bool linelet_format_decoder_parse_header(struct linelet_format_decoder_frame_params *params, const void *buffer_, size_t buffer_size)
{
   const uint8_t *orig_buffer = buffer_;
   const uint8_t *buffer = buffer_;

   if (memcmp(buffer, "LINELET1", 8))
      return false;
   buffer += 8;

   struct header_fmt1 fmt1 = {0};
   struct header_cod1 cod1 = {0};
   struct header_ord1 ord1 = { .pts = LINELET_FORMAT_PTS_INVALID };
   
   while ((size_t)(buffer + 8 - orig_buffer) < buffer_size)
   {
      if (!memcmp(buffer, "FMT1", 4))
      {
         buffer += 4;
         if (!(buffer = parse_fmt1(&fmt1, buffer)))
            return false;
      }
      else if (!memcmp(buffer, "COD1", 4))
      {
         buffer += 4;
         if (!(buffer = parse_cod1(&cod1, buffer)))
            return false;
      }
      else if (!memcmp(buffer, "ORD1", 4))
      {
         buffer += 4;
         if (!(buffer = parse_ord1(&ord1, buffer)))
            return false;
      }
      else if (!memcmp(buffer, "EOH1", 4))
      {
         uint32_t header_crc = linelet_checksum(orig_buffer, buffer - orig_buffer);
         uint32_t reference = linelet_read32_be(buffer + 4);

         if (reference != header_crc)
         {
            linelet_log_printf(LINELET_LOG_ERROR, "[Decoder]: Linelet header corrupt (CRC).\n");
            return false;
         }

         break;
      }
      else
      {
         // Skip.
         buffer += 4;
         buffer += linelet_read32_be(buffer);
      }
   }

   return header_to_frame_params(params, &fmt1, &cod1, &ord1);
}

static bool is_compatible_conversion(enum linelet_format_format target, enum linelet_format_format src)
{
   switch (target)
   {
      case LINELET_FMT_X2B10G10R10:
      case LINELET_FMT_GBRP10:
         return src == LINELET_FMT_YCBCR10_RCT_444;
      case LINELET_FMT_X8R8G8B8:
      case LINELET_FMT_GBRP8:
         return src == LINELET_FMT_YCBCR8_RCT_444;

      default:
         return target == src;
   }
}

bool linelet_format_decoder_process_frame_async(linelet_format_decoder_t *dec,
      const void *encoded_buffer_, size_t encoded_buffer_size,
      void * const *components, const size_t *component_strides)
{
   const uint8_t *encoded_buffer = encoded_buffer_;

   if (!parse_header(dec, encoded_buffer, encoded_buffer_size))
      return false;

   if (!is_compatible_conversion(dec->opts.format, dec->source_format))
      return false;

   unsigned height = (dec->fmt1.height + dec->cod1.slice_height - 1) & ~(dec->cod1.slice_height - 1);

   for (unsigned i = 0; i < dec->num_threads; i++)
   {
      struct thread_data *data = &dec->threads[i];

      unsigned start_y = (height * i) / dec->num_threads;
      unsigned end_y = (height * (i + 1)) / dec->num_threads;

      // Align to slice_height, except for last.
      start_y &= ~(dec->cod1.slice_height - 1);
      if (i < dec->num_threads - 1)
         end_y &= ~(dec->cod1.slice_height - 1);

      unsigned height = end_y - start_y;

      data->width = dec->fmt1.width;
      data->height = height;
      data->real_height = dec->fmt1.height;
      data->start_y = start_y;
      data->slice_height = dec->cod1.slice_height;
      data->postfilter = dec->cod1.postfilter;
      data->convert = format_to_convert_func(dec->opts.format);

      // Allocate YUV image and decoders for threads.
      if (data->img[0].width != dec->fmt1.width ||
            data->img[0].height != dec->cod1.slice_height ||
            (data->dec && (linelet_decoder_get_precinct_samples(data->dec) != dec->cod1.precinct_samples)))
      {
         for (unsigned j = 0; j < 2; j++)
         {
            if (!linelet_yuv_image_alloc(&data->img[j], dec->fmt1.width, dec->cod1.slice_height,
                     dec->cod1.components[0].decimation,
                     dec->cod1.components[1].decimation,
                     dec->cod1.components[2].decimation))
               return false;
         }

         linelet_decoder_free(data->dec);
         data->dec = linelet_decoder_new(dec->fmt1.width, dec->cod1.slice_height,
               dec->cod1.components, dec->cod1.num_components,
               &(const struct linelet_decoder_options) { .precinct_samples = dec->cod1.precinct_samples });
         if (!data->dec)
            return false;
      }
   }

   // Fire off threads.
   for (unsigned i = 0; i < dec->num_threads; i++)
   {
      struct thread_data *data = &dec->threads[i];

      data->payload = dec->payload;
      data->payload_end = dec->payload_end;

      data->scanline_meta = dec->scanline_meta + (data->start_y / dec->cod1.slice_height) * sizeof(uint32_t) * 2;

      for (unsigned c = 0; c < dec->num_components; c++)
      {
         data->dst[c] = (uint8_t*)components[c] + data->start_y * component_strides[c];
         data->stride[c] = component_strides[c];
      }

      pthread_mutex_lock(&data->lock);
      data->ready = true;
      pthread_cond_signal(&data->cond);
      pthread_mutex_unlock(&data->lock);
   }

   return true;
}

bool linelet_format_decoder_process_frame_wait(linelet_format_decoder_t *dec)
{
   // Wait until all threads are done processing.
   for (unsigned i = 0; i < dec->num_threads; i++)
   {
      struct thread_data *data = &dec->threads[i];
      pthread_mutex_lock(&data->lock);
      while (data->ready && !data->die)
         pthread_cond_wait(&data->cond, &data->lock);
      pthread_mutex_unlock(&data->lock);
   }

   return true;
}

bool linelet_format_decoder_process_frame(linelet_format_decoder_t *dec,
      const void *encoded_buffer, size_t encoded_buffer_size,
      void * const *components, const size_t *component_strides)
{
   if (linelet_format_decoder_process_frame_async(dec, encoded_buffer,
            encoded_buffer_size, components, component_strides))
      return linelet_format_decoder_process_frame_wait(dec);
   else
      return false;
}

