#include "pack.h"
#include "timer.h"
#include "memory.h"
#include "simd.h"
#include <stdlib.h>

size_t linelet_pack_predictive_unary(uint8_t *output, const uint8_t *input, unsigned samples)
{
   TIMER_COUNTER(pack_unary);
   TIMER_START_COUNT(pack_unary, samples);
#define clear(bit) output[(bit) >> 3] &= ~(1 << ((bit) & 7))
#define set(bit) output[(bit) >> 3] |= 1 << ((bit) & 7)
   unsigned bit = 0;
   int last = 0;

   for (unsigned i = 0; i < samples; i++)
   {
      int pred = input[i] - last;

      if (pred)
      {
         unsigned ones = abs(pred);
         for (unsigned b = 0; b < ones; b++, bit++)
            set(bit);
         clear(bit);
         bit++;

         if (last)
         {
            if (pred < 0)
               set(bit);
            else
               clear(bit);
            bit++;
         }
      }
      else
      {
         clear(bit);
         bit++;
      }

      last = input[i];
   }
#undef clear
#undef set

   TIMER_STOP(pack_unary);
   return (bit + 7) >> 3;
}

size_t linelet_unpack_predictive_unary(uint8_t *output, const uint8_t *input, unsigned samples)
{
   TIMER_COUNTER(unpack_unary);
   TIMER_START_COUNT(unpack_unary, samples);
#define isset(bit) (input[(bit) >> 3] & (1 << ((bit) & 7)))
   unsigned bit = 0;
   int last = 0;

   for (unsigned i = 0; i < samples; i++)
   {
      int val = 0;
      while (isset(bit))
      {
         val++;
         bit++;
      }

      bit++;
      if (val && last)
      {
         if (isset(bit))
            val = -val;
         bit++;
      }
      val += last;

      output[i] = last = val;
   }
#undef isset

   TIMER_STOP(unpack_unary);
   return (bit + 7) >> 3;
}

size_t linelet_pack_nibbles_u8(uint8_t *output, const uint8_t *input, unsigned samples)
{
   TIMER_COUNTER(pack_u8);
   TIMER_START_COUNT(pack_u8, samples);

   uint8_t *orig_output = output;

   size_t i;
   for (i = 0; i + 2 <= samples; i += 2)
      *output++ = ((input[i + 0] & 0xf) << 0) | ((input[i + 1] & 0xf) << 4);
   if (samples & 1)
      *output++ = input[i + 0] & 0xf;

   TIMER_STOP(pack_u8);
   return output - orig_output;
}

size_t linelet_unpack_nibbles_u8(uint8_t *output, const uint8_t *input, unsigned samples)
{
   TIMER_COUNTER(unpack_u8);
   TIMER_START_COUNT(unpack_u8, samples);

   size_t i;
   const uint8_t *orig_input = input;

   for (i = 0; i + 2 <= samples; i += 2)
   {
      uint8_t val = *input++;
      *output++ = (val >> 0) & 0xf;
      *output++ = (val >> 4) & 0xf;
   }

   if (samples & 1)
      *output++ = *input++ & 0xf;

   TIMER_STOP(unpack_u8);
   return input - orig_input;
}

#define PACK_2(bits) do { \
   uint32_t res = 0; \
   for (unsigned b = 0; b < bits; b++) \
   { \
      res <<= 2; \
      res |= ( \
       (((input[0] >> b) & 1) << 0) | \
       (((input[1] >> b) & 1) << 1)); \
   } \
   unsigned bytes = (bits + 3) >> 2; \
   for (unsigned b = 0; b < bytes; b++) \
   { \
      *output++ = res; \
      res >>= 8; \
   } \
} while(0)

size_t linelet_pack_bits_s16(uint8_t *output, const int16_t *input, unsigned samples, unsigned blocks, const uint8_t *bits)
{
   uint8_t *orig_output = output;

   if (samples >= 16)
   {
      TIMER_COUNTER(pack_16);
      TIMER_START_COUNT(pack_16, samples * blocks);
      const linelet_pack16_t *simd = linelet_simd.pack16;
      for (unsigned b = 0; b < blocks; b++)
      {
         unsigned bit = bits[b];
         if (bit)
            output = simd[bit - 1](output, input, samples);
         input += samples;
      }
      TIMER_STOP(pack_16);
   }
   else if (samples == 8)
   {
      TIMER_COUNTER(pack_8);
      TIMER_START_COUNT(pack_8, samples * blocks);
      const linelet_pack8_t *simd = linelet_simd.pack8;
      for (unsigned b = 0; b < blocks; b++, input += 8)
      {
         unsigned bit = bits[b];
         if (bit)
            output = simd[bit - 1](output, input);
      }
      TIMER_STOP(pack_8);
   }
   else if (samples == 4)
   {
      TIMER_COUNTER(pack_4);
      TIMER_START_COUNT(pack_4, samples * blocks);
      const linelet_pack4_t *simd = linelet_simd.pack4;
      for (unsigned b = 0; b < blocks; b++, input += 4)
      {
         unsigned bit = bits[b];
         if (bit)
            output = simd[bit - 1](output, input);
      }
      TIMER_STOP(pack_4);
   }
   else if (samples == 2)
   {
      TIMER_COUNTER(pack_2);
      TIMER_START_COUNT(pack_2, samples * blocks);
      for (unsigned b = 0; b < blocks; b++, input += 2)
      {
         switch (bits[b])
         {
            case  1: PACK_2(1); break;
            case  2: PACK_2(2); break;
            case  3: PACK_2(3); break;
            case  4: PACK_2(4); break;
            case  5: PACK_2(5); break;
            case  6: PACK_2(6); break;
            case  7: PACK_2(7); break;
            case  8: PACK_2(8); break;
            case  9: PACK_2(9); break;
            case 10: PACK_2(10); break;
            case 11: PACK_2(11); break;
            case 12: PACK_2(12); break;
            case 13: PACK_2(13); break;
            case 14: PACK_2(14); break;
            case 15: PACK_2(15); break;
            case 16: PACK_2(16); break;
         }
      }
      TIMER_STOP(pack_2);
   }

   // FIXME: Figure out what to do with pathological 1 sample case.

   linelet_simd.emms();
   return output - orig_output;
}

#define GEN_U64(a, b, c, d, e, f, g, h) ( \
      ((uint64_t)((a) & 1) << 0) | \
      ((uint64_t)((b) & 1) << 8) | \
      ((uint64_t)((c) & 1) << 16) | \
      ((uint64_t)((d) & 1) << 24) | \
      ((uint64_t)((e) & 1) << 32) | \
      ((uint64_t)((f) & 1) << 40) | \
      ((uint64_t)((g) & 1) << 48) | \
      ((uint64_t)((h) & 1) << 56))

#define C(x) GEN_U64((x) >> 0, (x) >> 1, (x) >> 2, (x) >> 3, (x) >> 4, (x) >> 5, (x) >> 6, (x) >> 7)
#define CODES(c) \
   C(c + 0), C(c + 1), C(c + 2), C(c + 3), C(c + 4), C(c + 5), C(c + 6), C(c + 7), \
   C(c + 8), C(c + 9), C(c + 10), C(c + 11), C(c + 12), C(c + 13), C(c + 14), C(c + 15)
static const uint64_t linelet_unpack_lut[256] = {
   CODES(0), CODES(16), CODES(32), CODES(48),
   CODES(64), CODES(80), CODES(96), CODES(112),
   CODES(128 + 0), CODES(128 + 16), CODES(128 + 32), CODES(128 + 48),
   CODES(128 + 64), CODES(128 + 80), CODES(128 + 96), CODES(128 + 112),
};
#undef GEN_U64
#undef C
#undef CODES

size_t linelet_unpack_bits_s16(int16_t *output, const uint8_t *input, unsigned samples, unsigned blocks, const uint8_t *bits)
{
   const uint8_t *orig_input = input;

   if (samples >= 8)
   {
      TIMER_COUNTER(unpack_s16_0);
      TIMER_START_COUNT(unpack_s16_0, samples * blocks);

      const linelet_unpack8_t *simd = linelet_simd.unpack8;
      for (unsigned b = 0; b < blocks; b++)
      {
         unsigned bit = bits[b];
         if (bit)
            input = simd[bit - 1](output, input, samples, linelet_unpack_lut);
         else
            memset(output, 0, samples * sizeof(int16_t));
         output += samples;
      }

      TIMER_STOP(unpack_s16_0);
   }
   else if (samples == 4)
   {
      TIMER_COUNTER(unpack_s16_2);
      TIMER_START_COUNT(unpack_s16_2, samples * blocks);

      const linelet_unpack4_t *simd = linelet_simd.unpack4;
      for (unsigned b = 0; b < blocks; b++, output += 4)
      {
         unsigned bit = bits[b];
         if (bit)
            input = simd[bit - 1](output, input, linelet_unpack_lut);
         else
            memset(output, 0, 4 * sizeof(int16_t));
      }

      TIMER_STOP(unpack_s16_2);
   }
   else if (samples == 2)
   {
      TIMER_COUNTER(unpack_s16_3);
      TIMER_START_COUNT(unpack_s16_3, samples * blocks);

      for (unsigned b = 0; b < blocks; b++, output += 2)
      {
         uint32_t res = 0;
         unsigned bits_tmp = bits[b];

         if (!bits_tmp)
         {
            memset(output, 0, samples * sizeof(int16_t));
            continue;
         }

         for (unsigned b = 0; b < bits_tmp; b++)
         {
            uint8_t v = (input[b >> 2] >> ((b & 3) * 2)) & 3;
            static const uint32_t splat_mask =  (1ull << 0) | (1ull << 15);
            static const uint32_t mask = 0x00010001;
            res |= ((v * splat_mask) & mask) << (15 - b);
         }

         for (unsigned s = 0; s < 2; s++)
         {
            uint16_t v = res;
            res >>= 16;
            output[s] = ((int16_t)v) >> (16 - bits_tmp);
         }

         input += (bits_tmp + 3) >> 2;
      }

      TIMER_STOP(unpack_s16_3);
   }

   // FIXME: Figure out what to do with pathological 1 sample case.

   linelet_simd.emms();
   return input - orig_input;
}

