#include "timer.h"
#include "log.h"
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errno.h>
#include <time.h>

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#ifndef TIMER_ATOMIC
#include <pthread.h>
static pthread_mutex_t register_lock = PTHREAD_MUTEX_INITIALIZER;
#endif

#define MAX_COUNTERS 128
static struct linelet_timer_counter *counters[MAX_COUNTERS];
static int num_counters;

int64_t linelet_timer_get_nsec(void)
{
#ifdef _WIN32
   // Frequency is guaranteed to not change.
   static LARGE_INTEGER freq;
   if (!freq.QuadPart && !QueryPerformanceFrequency(&freq))
      return 0;

   LARGE_INTEGER count;
   if (!QueryPerformanceCounter(&count))
      return 0;
   return count.QuadPart * 1000000000LL / freq.QuadPart;
#elif _POSIX_C_SOURCE >= 199307L
   struct timespec tv;
   clock_gettime(CLOCK_MONOTONIC, &tv);
   return tv.tv_sec * 1000000000LL + tv.tv_nsec;
#else
#error "No high-precicion clock implementation found."
#endif
}

int64_t linelet_timer_get_tick(void)
{
#if defined(__i386__) || defined(__i486__) || defined(__i686__)
   int64_t time;
   asm volatile (
         "xor %%eax, %%eax\n\t"
         "cpuid\n\t"
         "rdtsc\n\t"
         : "=A" (time) :: "ebx", "ecx", "memory");
   return time;
#elif defined(__x86_64__)
   unsigned a, d;
   asm volatile (
         "xor %%eax, %%eax\n\t"
         "cpuid\n\t"
         "rdtsc\n\t"
         : "=a" (a), "=d" (d) :: "ebx", "ecx", "memory");
   return (int64_t)a | ((int64_t)d << 32);
#else
   return linelet_timer_get_nsec();
#endif
}

void linelet_timer_sleep_until(int64_t target)
{
#ifdef _WIN32
   int64_t msecs = (target - linelet_timer_get_nsec() + 500000) / 1000000;
   if (msecs > 0)
      Sleep(msecs);
#elif defined(__USE_XOPEN2K)
   struct timespec tv = {
      .tv_sec = target / 1000000000,
      .tv_nsec = target % 1000000000,
   };

   if (clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME,
         &tv, NULL) < 0 && errno != EINTR)
         linelet_log_printf(LINELET_LOG_ERROR, "clock_nanosleep() failed.\n");
#elif _POSIX_C_SOURCE >= 199309L
   int64_t nsecs = target - linelet_timer_get_nsec();
   if (nsecs > 0)
   {
      struct timespec tv = {
         .tv_sec = nsecs / 1000000000,
         .tv_nsec = nsecs % 1000000000,
      };

      if (nanosleep(&tv, NULL) < 0 && errno != EINTR)
         linelet_log_printf(LINELET_LOG_ERROR, "nanosleep() failed.\n");

   }
#else
#error "No suitable implementation for linelet_timer_sleep_until() found."
#endif
}

#ifdef TIMER_ATOMIC
void linelet_timer_register(struct linelet_timer_counter *cnt)
{
   int num = __atomic_fetch_add(&num_counters, 1, __ATOMIC_SEQ_CST);
   if (num >= MAX_COUNTERS)
      return;

   cnt->registered = true;
   counters[num] = cnt;
}
#else
void linelet_timer_register(struct linelet_timer_counter *cnt)
{
   pthread_mutex_lock(&register_lock);
   if (num_counters >= MAX_COUNTERS)
      goto end;

   cnt->registered = true;
   counters[num_counters++] = cnt;

end:
   pthread_mutex_unlock(&register_lock);
}
#endif

static int counter_cmp(const void *a_, const void *b_)
{
   const struct linelet_timer_counter * const *a = a_;
   const struct linelet_timer_counter * const *b = b_;
   return strcmp((*a)->desc, (*b)->desc);
}

void linelet_timer_log(void)
{
#ifdef TIMER_ATOMIC
   int cnt = __atomic_load_n(&num_counters, __ATOMIC_SEQ_CST);
#else
   int cnt = num_counters;
#endif

   qsort(counters, num_counters, sizeof(*counters), counter_cmp);

   linelet_log_printf(LINELET_LOG_INFO, "[PERF]\n");
   for (int i = 0; i < cnt; i++)
      linelet_log_printf(LINELET_LOG_INFO, "%30s - MTicks: %10.3f, Ticks / Count: %18.3f, Count: %15"PRIu64"\n",
            counters[i]->desc, counters[i]->total / 1000000.0, (double)counters[i]->total / counters[i]->call_cnt, counters[i]->call_cnt);
}

