#include "rpng.h"
#include "log.h"

#include <zlib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static void *(*rpng_malloc)(size_t size) = malloc;
static void *(*rpng_calloc)(size_t memb_size, size_t membs) = calloc;
static void (*rpng_free)(void *ptr) = free;

void rpng_set_malloc(void *(*cb)(size_t))
{
   rpng_malloc = cb;
}

void rpng_set_calloc(void *(*cb)(size_t, size_t))
{
   rpng_calloc = cb;
}

void rpng_set_free(void (*cb)(void*))
{
   rpng_free = cb;
}

static inline uint32_t crc32_calculate(const uint8_t *data, size_t length)
{
   return crc32(0, data, length);
}

#undef GOTO_END_ERROR
#define GOTO_END_ERROR() do { \
   linelet_log_printf(LINELET_LOG_INFO, "[RPNG]: Error in line %d.\n", __LINE__); \
   ret = false; \
   goto end; \
} while(0)

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))
#endif

enum rpng_pixel_format
{
   RPNG_FMT_ARGB8,
   RPNG_FMT_ARGB16,
   RPNG_FMT_A2BGR10,
   RPNG_FMT_BGR24
};

static unsigned pixel_format_size(enum rpng_pixel_format fmt)
{
   switch (fmt)
   {
      case RPNG_FMT_ARGB8:
      case RPNG_FMT_A2BGR10:
         return sizeof(uint32_t);
      case RPNG_FMT_ARGB16:
         return sizeof(uint64_t);
      case RPNG_FMT_BGR24:
         return 3;
      default:
         return 0;
   }
}

static const uint8_t png_magic[8] = {
   0x89, 'P', 'N', 'G', 0x0d, 0x0a, 0x1a, 0x0a,
};

struct png_chunk
{
   uint32_t size;
   char type[4];
   uint8_t *data;
};

struct png_ihdr
{
   uint32_t width;
   uint32_t height;
   uint8_t depth;
   uint8_t color_type;
   uint8_t compression;
   uint8_t filter;
   uint8_t interlace;
};

enum png_chunk_type
{
   PNG_CHUNK_NOOP = 0,
   PNG_CHUNK_ERROR,
   PNG_CHUNK_IHDR,
   PNG_CHUNK_IDAT,
   PNG_CHUNK_PLTE,
   PNG_CHUNK_IEND
};

static uint32_t dword_be(const uint8_t *buf)
{
   return (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | (buf[3] << 0);
}

static bool read_chunk_header(FILE *file, struct png_chunk *chunk)
{
   uint8_t dword[4] = {0};
   if (fread(dword, 1, 4, file) != 4)
      return false;

   chunk->size = dword_be(dword);

   if (fread(chunk->type, 1, 4, file) != 4)
      return false;

   return true;
}

struct
{
   const char *id;
   enum png_chunk_type type;
} static const chunk_map[] = {
   { "IHDR", PNG_CHUNK_IHDR },
   { "IDAT", PNG_CHUNK_IDAT },
   { "IEND", PNG_CHUNK_IEND },
   { "PLTE", PNG_CHUNK_PLTE },
};

struct idat_buffer
{
   uint8_t *data;
   size_t size;
};

static enum png_chunk_type png_chunk_type(const struct png_chunk *chunk)
{
   unsigned i;
   for (i = 0; i < ARRAY_SIZE(chunk_map); i++)
   {
      if (memcmp(chunk->type, chunk_map[i].id, 4) == 0)
         return chunk_map[i].type;
   }

   return PNG_CHUNK_NOOP;
}

static bool png_read_chunk(FILE *file, struct png_chunk *chunk)
{
   rpng_free(chunk->data);
   chunk->data = (uint8_t*)rpng_calloc(1, chunk->size + sizeof(uint32_t)); // CRC32
   if (!chunk->data)
      return false;

   if (fread(chunk->data, 1, chunk->size + sizeof(uint32_t), file) != (chunk->size + sizeof(uint32_t)))
   {
      rpng_free(chunk->data);
      return false;
   }

   // Ignore CRC.
   return true;
}

static void png_rpng_free_chunk(struct png_chunk *chunk)
{
   rpng_free(chunk->data);
   chunk->data = NULL;
}

static bool png_parse_ihdr(FILE *file, struct png_chunk *chunk, struct png_ihdr *ihdr)
{
   unsigned i;
   bool ret = true;
   if (!png_read_chunk(file, chunk))
      return false;

   if (chunk->size != 13)
      GOTO_END_ERROR();

   ihdr->width       = dword_be(chunk->data + 0);
   ihdr->height      = dword_be(chunk->data + 4);
   ihdr->depth       = chunk->data[8];
   ihdr->color_type  = chunk->data[9];
   ihdr->compression = chunk->data[10];
   ihdr->filter      = chunk->data[11];
   ihdr->interlace   = chunk->data[12];

   if (ihdr->width == 0 || ihdr->height == 0)
      GOTO_END_ERROR();

   if (ihdr->color_type == 2 || ihdr->color_type == 4 || ihdr->color_type == 6)
   {
      if (ihdr->depth != 8 && ihdr->depth != 16)
         GOTO_END_ERROR();
   }
   else if (ihdr->color_type == 0)
   {
      static const unsigned valid_bpp[] = { 1, 2, 4, 8, 16 };
      bool correct_bpp = false;
      for (i = 0; i < ARRAY_SIZE(valid_bpp); i++)
      {
         if (valid_bpp[i] == ihdr->depth)
         {
            correct_bpp = true;
            break;
         }
      }

      if (!correct_bpp)
         GOTO_END_ERROR();
   }
   else if (ihdr->color_type == 3)
   {
      static const unsigned valid_bpp[] = { 1, 2, 4, 8 };
      bool correct_bpp = false;
      for (i = 0; i < ARRAY_SIZE(valid_bpp); i++)
      {
         if (valid_bpp[i] == ihdr->depth)
         {
            correct_bpp = true;
            break;
         }
      }

      if (!correct_bpp)
         GOTO_END_ERROR();
   }
   else
      GOTO_END_ERROR();

#ifdef RPNG_TEST
   linelet_log_printf(LINELET_LOG_INFO, "IHDR: (%u x %u), bpc = %u, palette = %s, color = %s, alpha = %s, adam7 = %s.\n",
         ihdr->width, ihdr->height,
         ihdr->depth, ihdr->color_type == 3 ? "yes" : "no",
         ihdr->color_type & 2 ? "yes" : "no",
         ihdr->color_type & 4 ? "yes" : "no",
         ihdr->interlace == 1 ? "yes" : "no");
#endif

   if (ihdr->compression != 0)
      GOTO_END_ERROR();

end:
   png_rpng_free_chunk(chunk);
   return ret;
}

// Paeth prediction filter.
static inline int paeth(int a, int b, int c)
{
   int p = a + b - c;
   int pa = abs(p - a);
   int pb = abs(p - b);
   int pc = abs(p - c);

   if (pa <= pb && pa <= pc)
      return a;
   else if (pb <= pc)
      return b;
   else
      return c;
}

// 16-bit reading routines.
static void copy_line16_rgb(uint8_t *data_, const uint8_t *decoded,
      unsigned width, unsigned bpp, const uint32_t *palette)
{
   uint64_t *data = (uint64_t*)data_;
   (void)palette;

   unsigned i;
   bpp /= 8;
   unsigned off = bpp - 1;
   // Either repeat the 8-bit bit pattern or get full 16-bit.
   for (i = 0; i < width; i++)
   {
      uint64_t r = (decoded[0] << 8) | (decoded[off]);
      decoded += bpp;
      uint64_t g = (decoded[0] << 8) | (decoded[off]);
      decoded += bpp;
      uint64_t b = (decoded[0] << 8) | (decoded[off]);
      decoded += bpp;
      data[i] = (0xffffull << 48) | (r << 32) | (g << 16) | (b << 0);
   }
}

static void copy_line16_rgba(uint8_t *data_, const uint8_t *decoded,
      unsigned width, unsigned bpp, const uint32_t *palette)
{
   uint64_t *data = (uint64_t*)data_;
   (void)palette;

   unsigned i;
   bpp /= 8;
   unsigned off = bpp - 1;
   // Either repeat the 8-bit bit pattern or get full 16-bit.
   for (i = 0; i < width; i++)
   {
      uint64_t r = (decoded[0] << 8) | (decoded[off]);
      decoded += bpp;
      uint64_t g = (decoded[0] << 8) | (decoded[off]);
      decoded += bpp;
      uint64_t b = (decoded[0] << 8) | (decoded[off]);
      decoded += bpp;
      uint64_t a = (decoded[0] << 8) | (decoded[off]);
      decoded += bpp;
      data[i] = (a << 48) | (r << 32) | (g << 16) | (b << 0);
   }
}

static void copy_line16_bw(uint8_t *data_, const uint8_t *decoded, unsigned width, unsigned depth, const uint32_t *palette)
{
   uint64_t *data = (uint64_t*)data_;
   (void)palette;

   unsigned i;
   if (depth == 16)
   {
      for (i = 0; i < width; i++, decoded += 2)
      {
         uint64_t val = (decoded[0] << 8) | decoded[1];
         data[i] = (val * 0x000100010001ull) | (0xffffull << 48);
      }
   }
   else
   {
      static const unsigned mul_table[] = { 0, 0xffff, 0x5555, 0, 0x1111, 0, 0, 0, 0x0101 };
      unsigned mul = mul_table[depth];
      unsigned mask = (1 << depth) - 1;
      unsigned bit = 0;
      for (i = 0; i < width; i++, bit += depth)
      {
         unsigned byte = bit >> 3;
         unsigned val = decoded[byte] >> (8 - depth - (bit & 7));

         val &= mask;
         val *= mul;
         data[i] = (val * 0x000100010001ull) | (0xffffull << 48);
      }
   }
}

static void copy_line16_gray_alpha(uint8_t *data_, const uint8_t *decoded, unsigned width,
      unsigned bpp, const uint32_t *palette)
{
   uint64_t *data = (uint64_t*)data_;
   (void)palette;

   unsigned i;
   bpp /= 8;
   unsigned off = bpp - 1;
   for (i = 0; i < width; i++)
   {
      uint64_t gray = (decoded[0] << 8) | decoded[off];
      decoded += bpp;
      uint64_t alpha = (decoded[0] << 8) | decoded[off];
      decoded += bpp;

      data[i] = (gray * 0x000100010001ull) | (alpha << 48);
   }
}

static void copy_line16_plt(uint8_t *data_, const uint8_t *decoded, unsigned width, unsigned depth, const uint32_t *palette)
{
   uint64_t *data = (uint64_t*)data_;
   (void)palette;

   unsigned i;
   unsigned mask = (1 << depth) - 1;
   unsigned bit = 0;
   for (i = 0; i < width; i++, bit += depth)
   {
      unsigned byte = bit >> 3;
      unsigned val = decoded[byte] >> (8 - depth - (bit & 7));
      val &= mask;

      uint32_t p = palette[val];
      uint64_t r = (p >> 16) & 0xff;
      uint64_t g = (p >>  8) & 0xff;
      uint64_t b = (p >>  0) & 0xff;
      r = (r << 8) | r;
      g = (g << 8) | g;
      b = (b << 8) | b;
      data[i] = (0xffffull << 48) | (r << 32) | (g << 16) | (b << 0);
   }
}

// 10-bit reading routines.
static void copy_line10_rgb(uint8_t *data_, const uint8_t *decoded,
      unsigned width, unsigned bpp, const uint32_t *palette)
{
   uint32_t *data = (uint32_t*)data_;
   (void)palette;

   unsigned i;
   bpp /= 8;
   unsigned off = bpp - 1;
   // Either repeat the 8-bit bit pattern or get full 16-bit.
   for (i = 0; i < width; i++)
   {
      uint32_t r = ((decoded[0] << 8) | (decoded[off])) >> 6;
      decoded += bpp;
      uint32_t g = ((decoded[0] << 8) | (decoded[off])) >> 6;
      decoded += bpp;
      uint32_t b = ((decoded[0] << 8) | (decoded[off])) >> 6;
      decoded += bpp;

      *data++ = (0x3u << 30) | (b << 20) | (g << 10) | (r << 0);
   }
}

static void copy_line10_rgba(uint8_t *data_, const uint8_t *decoded,
      unsigned width, unsigned bpp, const uint32_t *palette)
{
   uint32_t *data = (uint32_t*)data_;
   (void)palette;

   unsigned i;
   bpp /= 8;
   unsigned off = bpp - 1;
   // Either repeat the 8-bit bit pattern or get full 16-bit.
   for (i = 0; i < width; i++)
   {
      uint32_t r = ((decoded[0] << 8) | (decoded[off])) >> 6;
      decoded += bpp;
      uint32_t g = ((decoded[0] << 8) | (decoded[off])) >> 6;
      decoded += bpp;
      uint32_t b = ((decoded[0] << 8) | (decoded[off])) >> 6;
      decoded += bpp;
      // Skip alpha.
      decoded += bpp;

      *data++ = (0x3u << 30) | (b << 20) | (g << 10) | (r << 0);
   }
}

static void copy_line10_bw(uint8_t *data_, const uint8_t *decoded, unsigned width, unsigned depth, const uint32_t *palette)
{
   uint32_t *data = (uint32_t*)data_;
   (void)palette;

   unsigned i;
   if (depth == 16)
   {
      for (i = 0; i < width; i++, decoded += 2)
      {
         uint32_t val = ((decoded[0] << 8) | decoded[1]) >> 6;
         *data++ = (0x3u << 30) | (val << 20) | (val << 10) | (val << 0);
      }
   }
   else
   {
      static const unsigned mul_table[] = { 0, 0xffff, 0x5555, 0, 0x1111, 0, 0, 0, 0x0101 };
      unsigned mul = mul_table[depth];
      unsigned mask = (1 << depth) - 1;
      unsigned bit = 0;
      for (i = 0; i < width; i++, bit += depth)
      {
         unsigned byte = bit >> 3;
         unsigned val = decoded[byte] >> (8 - depth - (bit & 7));

         val &= mask;
         val *= mul;
         val >>= 6;
         *data++ = (0x3u << 30) | (val << 20) | (val << 10) | (val << 0);
      }
   }
}

static void copy_line10_gray_alpha(uint8_t *data_, const uint8_t *decoded, unsigned width,
      unsigned bpp, const uint32_t *palette)
{
   uint32_t *data = (uint32_t*)data_;
   (void)palette;

   unsigned i;
   bpp /= 8;
   unsigned off = bpp - 1;
   for (i = 0; i < width; i++)
   {
      uint32_t gray = ((decoded[0] << 8) | decoded[off]) >> 6;
      decoded += bpp;
      // Skip alpha.
      decoded += bpp;
      *data++ = (0x3u << 30) | (gray << 20) | (gray << 10) | (gray << 0);
   }
}

static void copy_line10_plt(uint8_t *data_, const uint8_t *decoded, unsigned width, unsigned depth, const uint32_t *palette)
{
   uint32_t *data = (uint32_t*)data_;
   (void)palette;

   unsigned i;
   unsigned mask = (1 << depth) - 1;
   unsigned bit = 0;
   for (i = 0; i < width; i++, bit += depth)
   {
      unsigned byte = bit >> 3;
      unsigned val = decoded[byte] >> (8 - depth - (bit & 7));
      val &= mask;

      uint32_t p = palette[val];
      uint16_t r = (p >> 16) & 0xff;
      uint16_t g = (p >>  8) & 0xff;
      uint16_t b = (p >>  0) & 0xff;
      r = (r << 2) | (r >> 6);
      g = (g << 2) | (g >> 6);
      b = (b << 2) | (b >> 6);
      *data++ = (0x3u << 30) | (b << 20) | (g << 10) | (b << 0);
   }
}

// 8-bit reading routines.
static void copy_line_rgb(uint8_t *data_, const uint8_t *decoded, unsigned width, unsigned bpp, const uint32_t *palette)
{
   uint32_t *data = (uint32_t*)data_;
   (void)palette;

   unsigned i;
   bpp /= 8;
   for (i = 0; i < width; i++)
   {
      uint32_t r = *decoded;
      decoded += bpp;
      uint32_t g = *decoded;
      decoded += bpp;
      uint32_t b = *decoded;
      decoded += bpp;
      data[i] = (0xffu << 24) | (r << 16) | (g << 8) | (b << 0);
   }
}

static void copy_line_rgba(uint8_t *data_, const uint8_t *decoded, unsigned width, unsigned bpp, const uint32_t *palette)
{
   uint32_t *data = (uint32_t*)data_;
   (void)palette;

   unsigned i;
   bpp /= 8;
   for (i = 0; i < width; i++)
   {
      uint32_t r = *decoded;
      decoded += bpp;
      uint32_t g = *decoded;
      decoded += bpp;
      uint32_t b = *decoded;
      decoded += bpp;
      uint32_t a = *decoded;
      decoded += bpp;
      data[i] = (a << 24) | (r << 16) | (g << 8) | (b << 0);
   }
}

static void copy_line_bw(uint8_t *data_, const uint8_t *decoded, unsigned width, unsigned depth, const uint32_t *palette)
{
   uint32_t *data = (uint32_t*)data_;
   (void)palette;

   unsigned i;
   if (depth == 16)
   {
      for (i = 0; i < width; i++)
      {
         uint32_t val = decoded[i << 1];
         data[i] = (val * 0x010101) | (0xffu << 24);
      }
   }
   else
   {
      static const unsigned mul_table[] = { 0, 0xff, 0x55, 0, 0x11, 0, 0, 0, 0x01 };
      unsigned mul = mul_table[depth];
      unsigned mask = (1 << depth) - 1;
      unsigned bit = 0;
      for (i = 0; i < width; i++, bit += depth)
      {
         unsigned byte = bit >> 3;
         unsigned val = decoded[byte] >> (8 - depth - (bit & 7));

         val &= mask;
         val *= mul;
         data[i] = (val * 0x010101) | (0xffu << 24);
      }
   }
}

static void copy_line_gray_alpha(uint8_t *data_, const uint8_t *decoded, unsigned width,
      unsigned bpp, const uint32_t *palette)
{
   uint32_t *data = (uint32_t*)data_;
   (void)palette;

   unsigned i;
   bpp /= 8;
   for (i = 0; i < width; i++)
   {
      uint32_t gray = *decoded;
      decoded += bpp;
      uint32_t alpha = *decoded;
      decoded += bpp;

      data[i] = (gray * 0x010101) | (alpha << 24);
   }
}

static void copy_line_plt(uint8_t *data_, const uint8_t *decoded, unsigned width, unsigned depth, const uint32_t *palette)
{
   uint32_t *data = (uint32_t*)data_;
   (void)palette;

   unsigned i;
   unsigned mask = (1 << depth) - 1;
   unsigned bit = 0;
   for (i = 0; i < width; i++, bit += depth)
   {
      unsigned byte = bit >> 3;
      unsigned val = decoded[byte] >> (8 - depth - (bit & 7));
      val &= mask;
      data[i] = palette[val];
   }
}

// End copy routines.

static void png_pass_geom(const struct png_ihdr *ihdr,
      unsigned width, unsigned height,
      unsigned *bpp_out, unsigned *pitch_out, size_t *pass_size)
{
   unsigned bpp;
   unsigned pitch;
   switch (ihdr->color_type)
   {
      case 0:
         bpp = (ihdr->depth + 7) / 8;
         pitch = (ihdr->width * ihdr->depth + 7) / 8;
         break;

      case 2:
         bpp = (ihdr->depth * 3 + 7) / 8;
         pitch = (ihdr->width * ihdr->depth * 3 + 7) / 8;
         break;

      case 3:
         bpp = (ihdr->depth + 7) / 8;
         pitch = (ihdr->width * ihdr->depth + 7) / 8;
         break;

      case 4:
         bpp = (ihdr->depth * 2 + 7) / 8;
         pitch = (ihdr->width * ihdr->depth * 2 + 7) / 8;
         break;

      case 6:
         bpp = (ihdr->depth * 4 + 7) / 8;
         pitch = (ihdr->width * ihdr->depth * 4 + 7) / 8;
         break;

      default:
         bpp = 0;
         pitch = 0;
         break;
   }

   if (pass_size)
      *pass_size = (pitch + 1) * ihdr->height;
   if (bpp_out)
      *bpp_out = bpp;
   if (pitch_out)
      *pitch_out = pitch;
}


static bool png_reverse_filter(uint8_t *data, enum rpng_pixel_format fmt, const struct png_ihdr *ihdr,
      const uint8_t *inflate_buf, size_t inflate_buf_size, const uint32_t *palette)
{
   unsigned i, h;
   bool ret = true;

   unsigned bpp;
   unsigned pitch;
   size_t pass_size;
   png_pass_geom(ihdr, ihdr->width, ihdr->height, &bpp, &pitch, &pass_size);
   unsigned pixel_size = pixel_format_size(fmt);

   if (inflate_buf_size < pass_size)
      return false;

   uint8_t *prev_scanline    = (uint8_t*)rpng_calloc(1, pitch);
   uint8_t *decoded_scanline = (uint8_t*)rpng_calloc(1, pitch);

   if (!prev_scanline || !decoded_scanline)
      GOTO_END_ERROR();

   void (*copy_line_func)(uint8_t *data, const uint8_t *decoded,
         unsigned width, unsigned depth, const uint32_t *palette) = NULL;

   switch (fmt)
   {
      case RPNG_FMT_ARGB16:
         if (ihdr->color_type == 0)
            copy_line_func = copy_line16_bw;
         else if (ihdr->color_type == 2)
            copy_line_func = copy_line16_rgb;
         else if (ihdr->color_type == 3)
            copy_line_func = copy_line16_plt;
         else if (ihdr->color_type == 4)
            copy_line_func = copy_line16_gray_alpha;
         else if (ihdr->color_type == 6)
            copy_line_func = copy_line16_rgba;
         break;

      case RPNG_FMT_A2BGR10:
         if (ihdr->color_type == 0)
            copy_line_func = copy_line10_bw;
         else if (ihdr->color_type == 2)
            copy_line_func = copy_line10_rgb;
         else if (ihdr->color_type == 3)
            copy_line_func = copy_line10_plt;
         else if (ihdr->color_type == 4)
            copy_line_func = copy_line10_gray_alpha;
         else if (ihdr->color_type == 6)
            copy_line_func = copy_line10_rgba;
         break;

      case RPNG_FMT_ARGB8:
         if (ihdr->color_type == 0)
            copy_line_func = copy_line_bw;
         else if (ihdr->color_type == 2)
            copy_line_func = copy_line_rgb;
         else if (ihdr->color_type == 3)
            copy_line_func = copy_line_plt;
         else if (ihdr->color_type == 4)
            copy_line_func = copy_line_gray_alpha;
         else if (ihdr->color_type == 6)
            copy_line_func = copy_line_rgba;
         break;

      default:
         return false;
   }

   for (h = 0; h < ihdr->height;
         h++, inflate_buf += pitch, data += ihdr->width * pixel_size)
   {
      unsigned filter = *inflate_buf++;
      switch (filter)
      {
         case 0: // None
            memcpy(decoded_scanline, inflate_buf, pitch);
            break;

         case 1: // Sub
            for (i = 0; i < bpp; i++)
               decoded_scanline[i] = inflate_buf[i];
            for (i = bpp; i < pitch; i++)
               decoded_scanline[i] = decoded_scanline[i - bpp] + inflate_buf[i];
            break;

         case 2: // Up
            for (i = 0; i < pitch; i++)
               decoded_scanline[i] = prev_scanline[i] + inflate_buf[i];
            break;

         case 3: // Average
            for (i = 0; i < bpp; i++)
            {
               uint8_t avg = prev_scanline[i] >> 1;
               decoded_scanline[i] = avg + inflate_buf[i];
            }
            for (i = bpp; i < pitch; i++)
            {
               uint8_t avg = (decoded_scanline[i - bpp] + prev_scanline[i]) >> 1;
               decoded_scanline[i] = avg + inflate_buf[i];
            }
            break;

         case 4: // Paeth
            for (i = 0; i < bpp; i++)
               decoded_scanline[i] = paeth(0, prev_scanline[i], 0) + inflate_buf[i];
            for (i = bpp; i < pitch; i++)
               decoded_scanline[i] = paeth(decoded_scanline[i - bpp], prev_scanline[i], prev_scanline[i - bpp]) + inflate_buf[i];
            break;

         default:
            GOTO_END_ERROR();
      }

      copy_line_func(data, decoded_scanline, ihdr->width, ihdr->depth, palette);
      memcpy(prev_scanline, decoded_scanline, pitch);
   }

end:
   rpng_free(decoded_scanline);
   rpng_free(prev_scanline);
   return ret;
}

struct adam7_pass
{
   unsigned x;
   unsigned y;
   unsigned stride_x;
   unsigned stride_y;
};

static void deinterlace_pass16(uint64_t *data, const struct png_ihdr *ihdr,
      const uint64_t *input, unsigned pass_width, unsigned pass_height, const struct adam7_pass *pass)
{
   unsigned x, y;
   data += pass->y * ihdr->width + pass->x;
   for (y = 0; y < pass_height; y++, data += ihdr->width * pass->stride_y, input += pass_width)
   {
      uint64_t *out = data;
      for (x = 0; x < pass_width; x++, out += pass->stride_x)
         *out = input[x];
   }
}

static void deinterlace_pass(uint32_t *data, const struct png_ihdr *ihdr,
      const uint32_t *input, unsigned pass_width, unsigned pass_height, const struct adam7_pass *pass)
{
   unsigned x, y;
   data += pass->y * ihdr->width + pass->x;
   for (y = 0; y < pass_height; y++, data += ihdr->width * pass->stride_y, input += pass_width)
   {
      uint32_t *out = data;
      for (x = 0; x < pass_width; x++, out += pass->stride_x)
         *out = input[x];
   }
}

static bool png_reverse_filter_adam7(uint8_t *data, enum rpng_pixel_format fmt, const struct png_ihdr *ihdr,
      const uint8_t *inflate_buf, size_t inflate_buf_size, const uint32_t *palette)
{
   unsigned pass;
   static const struct adam7_pass passes[] = {
      { 0, 0, 8, 8 },
      { 4, 0, 8, 8 },
      { 0, 4, 4, 8 },
      { 2, 0, 4, 4 },
      { 0, 2, 2, 4 },
      { 1, 0, 2, 2 },
      { 0, 1, 1, 2 },
   };

   unsigned pixel_size = pixel_format_size(fmt);

   for (pass = 0; pass < ARRAY_SIZE(passes); pass++)
   {
      if (ihdr->width <= passes[pass].x || ihdr->height <= passes[pass].y) // Empty pass
         continue;

      unsigned pass_width  = (ihdr->width - passes[pass].x + passes[pass].stride_x - 1) / passes[pass].stride_x;
      unsigned pass_height = (ihdr->height - passes[pass].y + passes[pass].stride_y - 1) / passes[pass].stride_y;

      uint8_t *tmp_data = (uint8_t*)rpng_malloc(pass_width * pass_height * pixel_size);
      if (!tmp_data)
         return false;

      struct png_ihdr tmp_ihdr = *ihdr;
      tmp_ihdr.width = pass_width;
      tmp_ihdr.height = pass_height;

      size_t pass_size;
      png_pass_geom(&tmp_ihdr, pass_width, pass_height, NULL, NULL, &pass_size);

      if (pass_size > inflate_buf_size)
      {
         rpng_free(tmp_data);
         return false;
      }

      if (!png_reverse_filter(tmp_data, fmt, &tmp_ihdr, inflate_buf, pass_size, palette))
      {
         rpng_free(tmp_data);
         return false;
      }

      inflate_buf += pass_size;
      inflate_buf_size -= pass_size;

      if (pixel_size == sizeof(uint64_t))
         deinterlace_pass16((uint64_t*)data, ihdr, (const uint64_t*)tmp_data, pass_width, pass_height, &passes[pass]);
      else if (pixel_size == sizeof(uint32_t))
         deinterlace_pass((uint32_t*)data, ihdr, (const uint32_t*)tmp_data, pass_width, pass_height, &passes[pass]);

      rpng_free(tmp_data);
   }

   return true;
}

static bool png_append_idat(FILE *file, const struct png_chunk *chunk, struct idat_buffer *buf)
{
   uint8_t *new_buffer = (uint8_t*)realloc(buf->data, buf->size + chunk->size);
   if (!new_buffer)
      return false;

   buf->data  = new_buffer;
   if (fread(buf->data + buf->size, 1, chunk->size, file) != chunk->size)
      return false;
   if (fseek(file, sizeof(uint32_t), SEEK_CUR) < 0)
      return false;
   buf->size += chunk->size;
   return true;
}

static bool png_read_plte(FILE *file, uint32_t *buffer, unsigned entries)
{
   unsigned i;
   if (entries > 256)
      return false;

   uint8_t buf[256 * 3];
   if (fread(buf, 3, entries, file) != entries)
      return false;

   for (i = 0; i < entries; i++)
   {
      uint32_t r = buf[3 * i + 0];
      uint32_t g = buf[3 * i + 1];
      uint32_t b = buf[3 * i + 2];
      buffer[i] = (r << 16) | (g << 8) | (b << 0) | (0xffu << 24);
   }

   if (fseek(file, sizeof(uint32_t), SEEK_CUR) < 0)
      return false;

   return true;
}

static bool rpng_load_image(const char *path, uint8_t **data, unsigned *width, unsigned *height, enum rpng_pixel_format fmt)
{
   long pos;
   *data   = NULL;
   *width  = 0;
   *height = 0;

   bool ret = true;
   FILE *file = fopen(path, "rb");
   if (!file)
      return false;

   fseek(file, 0, SEEK_END);
   long file_len = ftell(file);
   rewind(file);

   bool has_ihdr = false;
   bool has_idat = false;
   bool has_iend = false;
   bool has_plte = false;
   uint8_t *inflate_buf = NULL;
   size_t inflate_buf_size = 0;
   z_stream stream = {0};

   struct idat_buffer idat_buf = {0};
   struct png_ihdr ihdr = {0};
   uint32_t palette[256] = {0};

   char header[8];
   if (fread(header, 1, sizeof(header), file) != sizeof(header))
      GOTO_END_ERROR();

   if (memcmp(header, png_magic, sizeof(png_magic)) != 0)
      GOTO_END_ERROR();

   // feof() apparently isn't triggered after a seek (IEND).
   for (pos = ftell(file); pos < file_len && pos >= 0; pos = ftell(file))
   {
      struct png_chunk chunk = {0};
      if (!read_chunk_header(file, &chunk))
         GOTO_END_ERROR();

      switch (png_chunk_type(&chunk))
      {
         case PNG_CHUNK_NOOP:
         default:
            if (fseek(file, chunk.size + sizeof(uint32_t), SEEK_CUR) < 0)
               GOTO_END_ERROR();
            break;

         case PNG_CHUNK_ERROR:
            GOTO_END_ERROR();

         case PNG_CHUNK_IHDR:
            if (has_ihdr || has_idat || has_iend)
               GOTO_END_ERROR();

            if (!png_parse_ihdr(file, &chunk, &ihdr))
               GOTO_END_ERROR();

            has_ihdr = true;
            break;

         case PNG_CHUNK_PLTE:
            if (!has_ihdr || has_plte || has_iend || has_idat)
               GOTO_END_ERROR();

            if (chunk.size % 3)
               GOTO_END_ERROR();

            if (!png_read_plte(file, palette, chunk.size / 3))
               GOTO_END_ERROR();

            has_plte = true;
            break;

         case PNG_CHUNK_IDAT:
            if (!has_ihdr || has_iend || (ihdr.color_type == 3 && !has_plte))
               GOTO_END_ERROR();

            if (!png_append_idat(file, &chunk, &idat_buf))
               GOTO_END_ERROR();

            has_idat = true;
            break;

         case PNG_CHUNK_IEND:
            if (!has_ihdr || !has_idat)
               GOTO_END_ERROR();

            if (fseek(file, sizeof(uint32_t), SEEK_CUR) < 0)
               GOTO_END_ERROR();

            has_iend = true;
            break;
      }
   }

   if (!has_ihdr || !has_idat || !has_iend)
      GOTO_END_ERROR();

   if (inflateInit(&stream) != Z_OK)
      GOTO_END_ERROR();

   png_pass_geom(&ihdr, ihdr.width, ihdr.height, NULL, NULL, &inflate_buf_size);
   if (ihdr.interlace == 1) // To be sure.
      inflate_buf_size *= 2;

   inflate_buf = (uint8_t*)rpng_malloc(inflate_buf_size);
   if (!inflate_buf)
      GOTO_END_ERROR();

   stream.next_in   = idat_buf.data;
   stream.avail_in  = idat_buf.size;
   stream.avail_out = inflate_buf_size;
   stream.next_out  = inflate_buf;

   if (inflate(&stream, Z_FINISH) != Z_STREAM_END)
   {
      inflateEnd(&stream);
      GOTO_END_ERROR();
   }
   inflateEnd(&stream);

   *width  = ihdr.width;
   *height = ihdr.height;
   *data = (uint8_t*)rpng_malloc(ihdr.width * ihdr.height * pixel_format_size(fmt));
   if (!*data)
      GOTO_END_ERROR();

   if (ihdr.interlace == 1)
   {
      if (!png_reverse_filter_adam7(*data, fmt, &ihdr, inflate_buf, stream.total_out, palette))
         GOTO_END_ERROR();
   }
   else if (!png_reverse_filter(*data, fmt, &ihdr, inflate_buf, stream.total_out, palette))
      GOTO_END_ERROR();

end:
   if (file)
      fclose(file);
   if (!ret)
      rpng_free(*data);
   free(idat_buf.data);
   rpng_free(inflate_buf);
   return ret;
}

bool rpng_load_image_argb16(const char *path, uint64_t **data, unsigned *width, unsigned *height)
{
   return rpng_load_image(path, (uint8_t**)data, width, height, RPNG_FMT_ARGB16);
}

bool rpng_load_image_a2bgr10(const char *path, uint32_t **data, unsigned *width, unsigned *height)
{
   return rpng_load_image(path, (uint8_t**)data, width, height, RPNG_FMT_A2BGR10);
}

bool rpng_load_image_argb(const char *path, uint32_t **data, unsigned *width, unsigned *height)
{
   return rpng_load_image(path, (uint8_t**)data, width, height, RPNG_FMT_ARGB8);
}

static void dword_write_be(uint8_t *buf, uint32_t val)
{
   *buf++ = (uint8_t)(val >> 24);
   *buf++ = (uint8_t)(val >> 16);
   *buf++ = (uint8_t)(val >>  8);
   *buf++ = (uint8_t)(val >>  0);
}

static bool png_write_crc(FILE *file, const uint8_t *data, size_t size)
{
   uint32_t crc = crc32_calculate(data, size);
   uint8_t crc_raw[4] = {0};
   dword_write_be(crc_raw, crc);
   return fwrite(crc_raw, 1, sizeof(crc_raw), file) == sizeof(crc_raw);
}

static bool png_write_ihdr(FILE *file, const struct png_ihdr *ihdr)
{
   uint8_t ihdr_raw[] = {
      '0', '0', '0', '0', // Size
      'I', 'H', 'D', 'R',

      0, 0, 0, 0, // Width
      0, 0, 0, 0, // Height
      ihdr->depth,
      ihdr->color_type,
      ihdr->compression,
      ihdr->filter,
      ihdr->interlace,
   };

   dword_write_be(ihdr_raw +  0, sizeof(ihdr_raw) - 8);
   dword_write_be(ihdr_raw +  8, ihdr->width);
   dword_write_be(ihdr_raw + 12, ihdr->height);
   if (fwrite(ihdr_raw, 1, sizeof(ihdr_raw), file) != sizeof(ihdr_raw))
      return false;

   if (!png_write_crc(file, ihdr_raw + sizeof(uint32_t), sizeof(ihdr_raw) - sizeof(uint32_t)))
      return false;

   return true;
}

static bool png_write_idat(FILE *file, const uint8_t *data, size_t size)
{
   if (fwrite(data, 1, size, file) != size)
      return false;

   if (!png_write_crc(file, data + sizeof(uint32_t), size - sizeof(uint32_t)))
      return false;

   return true;
}

static bool png_write_iend(FILE *file)
{
   const uint8_t data[] = {
      0, 0, 0, 0,
      'I', 'E', 'N', 'D',
   };

   if (fwrite(data, 1, sizeof(data), file) != sizeof(data))
      return false;

   if (!png_write_crc(file, data + sizeof(uint32_t), sizeof(data) - sizeof(uint32_t)))
      return false;

   return true;
}

static void copy_argb16_line(uint8_t *dst, const uint64_t *src, unsigned width)
{
   unsigned i;
   for (i = 0; i < width; i++)
   {
      uint64_t col = src[i];
      *dst++ = (uint8_t)(col >> 40);
      *dst++ = (uint8_t)(col >> 32);
      *dst++ = (uint8_t)(col >> 24);
      *dst++ = (uint8_t)(col >> 16);
      *dst++ = (uint8_t)(col >> 8);
      *dst++ = (uint8_t)(col >> 0);
      *dst++ = (uint8_t)(col >> 56);
      *dst++ = (uint8_t)(col >> 48);
   }
}

static void copy_a2bgr10_line(uint8_t *dst, const uint32_t *src, unsigned width)
{
   unsigned i;
   for (i = 0; i < width; i++)
   {
      uint32_t col = *src++;
      uint16_t r = (col >> 0) & 0x3ff;
      uint16_t g = (col >> 10) & 0x3ff;
      uint16_t b = (col >> 20) & 0x3ff;
      r = (r << 6) | (r >> 4);
      g = (g << 6) | (g >> 4);
      b = (b << 6) | (b >> 4);

      *dst++ = (uint8_t)(r >> 8);
      *dst++ = (uint8_t)(r >> 0);
      *dst++ = (uint8_t)(g >> 8);
      *dst++ = (uint8_t)(g >> 0);
      *dst++ = (uint8_t)(b >> 8);
      *dst++ = (uint8_t)(b >> 0);
   }
}

static void copy_argb_line(uint8_t *dst, const uint32_t *src, unsigned width)
{
   unsigned i;
   for (i = 0; i < width; i++)
   {
      uint32_t col = src[i];
      *dst++ = (uint8_t)(col >> 16);
      *dst++ = (uint8_t)(col >>  8);
      *dst++ = (uint8_t)(col >>  0);
      *dst++ = (uint8_t)(col >> 24);
   }
}

static void copy_bgr24_line(uint8_t *dst, const uint8_t *src, unsigned width)
{
   unsigned i;
   for (i = 0; i < width; i++, dst += 3, src += 3)
   {
      dst[2] = src[0];
      dst[1] = src[1];
      dst[0] = src[2];
   }
}

static unsigned count_sad(const uint8_t *data, size_t size)
{
   size_t i;
   unsigned cnt = 0;
   for (i = 0; i < size; i++)
      cnt += abs((int8_t)data[i]);
   return cnt;
}

static unsigned filter_up(uint8_t *target, const uint8_t *line, const uint8_t *prev,
      unsigned width, unsigned bpp)
{
   unsigned i;
   width *= bpp;
   for (i = 0; i < width; i++)
      target[i] = line[i] - prev[i];

   return count_sad(target, width);
}

static unsigned filter_sub(uint8_t *target, const uint8_t *line,
      unsigned width, unsigned bpp)
{
   unsigned i;
   width *= bpp;
   for (i = 0; i < bpp; i++)
      target[i] = line[i];
   for (i = bpp; i < width; i++)
      target[i] = line[i] - line[i - bpp];

   return count_sad(target, width);
}

static unsigned filter_avg(uint8_t *target, const uint8_t *line, const uint8_t *prev,
      unsigned width, unsigned bpp)
{
   unsigned i;
   width *= bpp;
   for (i = 0; i < bpp; i++)
      target[i] = line[i] - (prev[i] >> 1);
   for (i = bpp; i < width; i++)
      target[i] = line[i] - ((line[i - bpp] + prev[i]) >> 1);

   return count_sad(target, width);
}

static unsigned filter_paeth(uint8_t *target, const uint8_t *line, const uint8_t *prev,
      unsigned width, unsigned bpp)
{
   unsigned i;
   width *= bpp;
   for (i = 0; i < bpp; i++)
      target[i] = line[i] - paeth(0, prev[i], 0);
   for (i = bpp; i < width; i++)
      target[i] = line[i] - paeth(line[i - bpp], prev[i], prev[i - bpp]);

   return count_sad(target, width);
}

static bool rpng_save_image(const char *path, const uint8_t *data,
      unsigned width, unsigned height, unsigned pitch, enum rpng_pixel_format fmt, unsigned compression)
{
   unsigned h;
   bool ret = true;
   struct png_ihdr ihdr = {0};

   size_t encode_buf_size  = 0;
   uint8_t *encode_buf     = NULL;
   uint8_t *deflate_buf    = NULL;
   uint8_t *rgba_line      = NULL;
   uint8_t *up_filtered    = NULL;
   uint8_t *sub_filtered   = NULL;
   uint8_t *avg_filtered   = NULL;
   uint8_t *paeth_filtered = NULL;
   uint8_t *prev_encoded   = NULL;
   uint8_t *encode_target  = NULL;

   z_stream stream = {0};

   FILE *file = fopen(path, "wb");
   if (!file)
      GOTO_END_ERROR();

   if (fwrite(png_magic, 1, sizeof(png_magic), file) != sizeof(png_magic))
      GOTO_END_ERROR();

   ihdr.width = width;
   ihdr.height = height;

   unsigned bpp;
   switch (fmt)
   {
      case RPNG_FMT_ARGB8:
         ihdr.depth = 8;
         ihdr.color_type = 6;
         bpp = sizeof(uint32_t);
         break;

      case RPNG_FMT_ARGB16:
         ihdr.depth = 16;
         ihdr.color_type = 6;
         bpp = sizeof(uint64_t);
         break;

      case RPNG_FMT_A2BGR10:
         ihdr.depth = 16;
         ihdr.color_type = 2;
         bpp = sizeof(uint16_t) * 3;
         break;

      case RPNG_FMT_BGR24:
         ihdr.depth = 8;
         ihdr.color_type = 2;
         bpp = 3;
         break;

      default:
         GOTO_END_ERROR();
   }

   if (!png_write_ihdr(file, &ihdr))
      GOTO_END_ERROR();

   encode_buf_size = (width * bpp + 1) * height;
   encode_buf = (uint8_t*)rpng_malloc(encode_buf_size);
   if (!encode_buf)
      GOTO_END_ERROR();

   prev_encoded = (uint8_t*)rpng_calloc(1, width * bpp);
   if (!prev_encoded)
      GOTO_END_ERROR();

   rgba_line      = (uint8_t*)rpng_malloc(width * bpp);
   up_filtered    = (uint8_t*)rpng_malloc(width * bpp);
   sub_filtered   = (uint8_t*)rpng_malloc(width * bpp);
   avg_filtered   = (uint8_t*)rpng_malloc(width * bpp);
   paeth_filtered = (uint8_t*)rpng_malloc(width * bpp);
   if (!rgba_line || !up_filtered || !sub_filtered || !avg_filtered || !paeth_filtered)
      GOTO_END_ERROR();

   encode_target = encode_buf;
   for (h = 0; h < height;
         h++, encode_target += width * bpp, data += pitch)
   {
      switch (fmt)
      {
         case RPNG_FMT_ARGB16:
            copy_argb16_line(rgba_line, (const uint64_t*)data, width);
            break;

         case RPNG_FMT_A2BGR10:
            copy_a2bgr10_line(rgba_line, (const uint32_t*)data, width);
            break;

         case RPNG_FMT_ARGB8:
            copy_argb_line(rgba_line, (const uint32_t*)data, width);
            break;

         case RPNG_FMT_BGR24:
            copy_bgr24_line(rgba_line, data, width);
            break;

         default:
            GOTO_END_ERROR();
      }

      const uint8_t *chosen_filtered = rgba_line;
      uint8_t filter = 0;

      if (compression)
      {
         // Try every filtering method, and choose the method
         // which gives minimum sum-absolute-difference (SAD).
         // This is probably not very optimal considering zlib compression is used,
         // but it's very simple to implement.
         unsigned none_score  = count_sad(rgba_line, width * bpp);
         unsigned up_score    = filter_up(up_filtered, rgba_line, prev_encoded, width, bpp);
         unsigned sub_score   = filter_sub(sub_filtered, rgba_line, width, bpp);
         unsigned avg_score   = filter_avg(avg_filtered, rgba_line, prev_encoded, width, bpp);
         unsigned paeth_score = filter_paeth(paeth_filtered, rgba_line, prev_encoded, width, bpp);

         unsigned min_sad = none_score;

         if (sub_score < min_sad)
         {
            filter = 1;
            chosen_filtered = sub_filtered;
            min_sad = sub_score;
         }

         if (up_score < min_sad)
         {
            filter = 2;
            chosen_filtered = up_filtered;
            min_sad = up_score;
         }

         if (avg_score < min_sad)
         {
            filter = 3;
            chosen_filtered = avg_filtered;
            min_sad = avg_score;
         }

         if (paeth_score < min_sad)
         {
            filter = 4;
            chosen_filtered = paeth_filtered;
            min_sad = paeth_score;
         }
      }

      *encode_target++ = filter;
      memcpy(encode_target, chosen_filtered, width * bpp);

      memcpy(prev_encoded, rgba_line, width * bpp);
   }

   deflate_buf = (uint8_t*)rpng_malloc(encode_buf_size * 2); // Just to be sure.
   if (!deflate_buf)
      GOTO_END_ERROR();

   stream.next_in   = encode_buf;
   stream.avail_in  = encode_buf_size;
   stream.next_out  = deflate_buf + 8;
   stream.avail_out = encode_buf_size * 2;

   deflateInit(&stream, compression);
   if (deflate(&stream, Z_FINISH) != Z_STREAM_END)
   {
      deflateEnd(&stream);
      GOTO_END_ERROR();
   }
   deflateEnd(&stream);

   memcpy(deflate_buf + 4, "IDAT", 4);
   dword_write_be(deflate_buf + 0, stream.total_out);
   if (!png_write_idat(file, deflate_buf, stream.total_out + 8))
      GOTO_END_ERROR();

   if (!png_write_iend(file))
      GOTO_END_ERROR();

end:
   if (file)
      fclose(file);
   rpng_free(encode_buf);
   rpng_free(deflate_buf);
   rpng_free(rgba_line);
   rpng_free(prev_encoded);
   rpng_free(up_filtered);
   rpng_free(sub_filtered);
   rpng_free(avg_filtered);
   rpng_free(paeth_filtered);
   return ret;
}

bool rpng_save_image_argb16(const char *path, const uint64_t *data,
      unsigned width, unsigned height, unsigned pitch, unsigned compression)
{
   return rpng_save_image(path, (const uint8_t*)data, width, height, pitch,
         RPNG_FMT_ARGB16, compression);
}

bool rpng_save_image_a2bgr10(const char *path, const uint32_t *data,
      unsigned width, unsigned height, unsigned pitch, unsigned compression)
{
   return rpng_save_image(path, (const uint8_t*)data, width, height, pitch,
         RPNG_FMT_A2BGR10, compression);
}

bool rpng_save_image_argb(const char *path, const uint32_t *data,
      unsigned width, unsigned height, unsigned pitch, unsigned compression)
{
   return rpng_save_image(path, (const uint8_t*)data, width, height, pitch,
         RPNG_FMT_ARGB8, compression);
}

bool rpng_save_image_bgr24(const char *path, const uint8_t *data,
      unsigned width, unsigned height, unsigned pitch, unsigned compression)
{
   return rpng_save_image(path, (const uint8_t*)data, width, height, pitch,
         RPNG_FMT_BGR24, compression);
}

