#include "tiff.h"
#include "log.h"
#include "memory.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Trivial. Decodes just what is needed for StEM material.
//

static uint16_t read_word_le(const uint8_t *buf) { return buf[0] | (buf[1] << 8); }
static uint16_t read_word_be(const uint8_t *buf) { return (buf[0] << 8) | buf[1]; }
static uint32_t read_dword_le(const uint8_t *buf) { return buf[0] | (buf[1] << 8) | (buf[2] << 16) | (buf[3] << 24); }
static uint32_t read_dword_be(const uint8_t *buf) { return (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3]; }

static inline bool is_little_endian(void)
{
   uint16_t val = 1;
   uint8_t test[2];
   memcpy(test, &val, sizeof(val));
   return test[0];
}

bool linelet_tiff_load_image_a2bgr10(const char *path, uint32_t **pixels,
      unsigned *out_width, unsigned *out_height)
{
   *pixels = NULL;
   *out_width = 0;
   *out_height = 0;

   FILE *file = fopen(path, "rb");
   if (!file)
      return false;

   fseek(file, 0, SEEK_END);
   long file_len = ftell(file);
   rewind(file);

   uint32_t *strip_offsets = NULL;
   uint8_t *buffer = linelet_malloc(file_len);
   if (!buffer)
      goto error;

   if (fread(buffer, 1, file_len, file) != (size_t)file_len)
      goto error;

   fclose(file);
   file = NULL;

   uint16_t (*read_word)(const uint8_t *buf) = NULL;
   uint32_t (*read_dword)(const uint8_t *buf) = NULL;

   bool is_native_endian;

   if (memcmp(buffer, "II", 2) == 0) // Little-endian
   {
      read_word = read_word_le;
      read_dword = read_dword_le;
      is_native_endian = is_little_endian();
   }
   else if (memcmp(buffer, "MM", 2) == 0) // Big-endian
   {
      read_word = read_word_be;
      read_dword = read_dword_be;
      is_native_endian = !is_little_endian();
   }
   else
      goto error;

   if (read_word(buffer + 2) != 42)
      goto error;

   uint32_t off = read_dword(buffer + 4);
   unsigned entries = read_word(buffer + off);

   unsigned width = 1;
   unsigned height = 1;
   unsigned bits[3] = {1, 1, 1};
   unsigned samples = 1;
   unsigned compression = 1;
   unsigned planar_conf = 1;
   bool is_rgb = false;

   uint32_t rows_per_strip = 0xffffffffu;
   unsigned num_strips = 0;

   while (off)
   {
      for (unsigned i = 0; i < entries; i++)
      {
         uint16_t tag = read_word(buffer + off + 2);
         uint16_t type = read_word(buffer + off + 4);
         uint32_t len = read_dword(buffer + off + 6);

         uint32_t data_word = read_word(buffer + off + 10);
         uint32_t data_dword = read_dword(buffer + off + 10);
         uint32_t offset = data_dword;

         switch (tag)
         {
            case 258: // BitsPerSample
               if (len == 3)
               {
                  bits[0] = read_word(buffer + offset);
                  bits[1] = read_word(buffer + offset + 2);
                  bits[2] = read_word(buffer + offset + 4);
               }
               break;
            case 257: // ImageLength
               height = type == 4 ? data_dword : data_word;
               break;
            case 256: // ImageWidth
               width = type == 4 ? data_dword : data_word;
               break;
            case 262: // PhotometricInterpretation
               is_rgb = data_word == 2;
               break;
            case 277: // SamplesPerPixel
               samples = data_word;
               break;
            case 259: // Compression
               compression = data_word;
               break;
            case 284: // PlanarConfiguration
               planar_conf = data_word;
               break;
            case 273: // StripOffsets
               strip_offsets = linelet_calloc(len, sizeof(uint32_t));
               num_strips = len;
               if (!strip_offsets)
                  goto error;
               if (type == 4)
               {
                  for (uint32_t n = 0; n < len; n++)
                     strip_offsets[n] = read_dword(buffer + offset + n * 4);
               }
               else
               {
                  for (uint32_t n = 0; n < len; n++)
                     strip_offsets[n] = read_word(buffer + offset + n * 2);
               }
               break;
            case 278: // RowsPerStrip
               rows_per_strip = type == 4 ? data_dword : data_word;
               break;
            default: // ?!
               break;
         }

         off += 12;
      }

      off = read_dword(buffer + off + 2);
   }

   if (bits[0] != 16 || bits[1] != 16 || bits[2] != 16)
      goto error;
   if (samples != 3)
      goto error;
   if (compression != 1)
      goto error;
   if (planar_conf != 1)
      goto error;
   if (!strip_offsets)
      goto error;
   if (!is_rgb)
      goto error;

   if (rows_per_strip > height)
      rows_per_strip = height;

   linelet_log_printf(LINELET_LOG_INFO, "[TIFF]: (%u x %u)\n", width, height);

   *pixels = linelet_malloc(width * height * sizeof(uint32_t));
   *out_width = width;
   *out_height = height;

   uint32_t *target = *pixels;

   if (is_native_endian)
   {
      for (unsigned s = 0; s < num_strips; s++)
      {
         for (unsigned y = 0; y < rows_per_strip; y++)
         {
            const uint16_t *src = (const uint16_t*)(buffer + strip_offsets[s]);
            for (unsigned x = 0; x < width; x++)
            {
               uint16_t r = *src++ >> 6;
               uint16_t g = *src++ >> 6;
               uint16_t b = *src++ >> 6;
               *target++ = (0x3u << 30) | (b << 20) | (g << 10) | (r << 0);
            }
         }
      }
   }
   else
   {
      for (unsigned s = 0; s < num_strips; s++)
      {
         for (unsigned y = 0; y < rows_per_strip; y++)
         {
            const uint8_t *src = buffer + strip_offsets[s];
            for (unsigned x = 0; x < width; x++, src += 6)
            {
               uint16_t r = read_word(src + 0) >> 6;
               uint16_t g = read_word(src + 2) >> 6;
               uint16_t b = read_word(src + 4) >> 6;
               *target++ = (0x3u << 30) | (b << 20) | (g << 10) | (r << 0);
            }
         }
      }
   }

   linelet_free(buffer);
   linelet_free(strip_offsets);
   return true;

error:
   if (file)
      fclose(file);
   linelet_free(*pixels);
   linelet_free(strip_offsets);
   *pixels = NULL;
   *out_width = *out_height = 0;
   return false;
}

