#include "tests.h"
#include "memory.h"
#include <stdio.h>
#include <stdlib.h>

int encoder_range_test(void)
{
   int ret = 0;
   linelet_encoder_t *enc = NULL;
   linelet_decoder_t *dec = NULL;

   unsigned width = 128 * 1024;
   unsigned height = 8;

   int16_t *component = NULL;
   int16_t *reference = NULL;
   int16_t *out_component = NULL;
   uint8_t *encoded_buffer = NULL;

   enc = linelet_encoder_new(width, height,
         &(const struct linelet_component) {
            .decomposition_levels = 7,
            .decimation = 0,
         }, 1,
         &(const struct linelet_encoder_options) {
            .precinct_samples = 256,
         });

   if (!enc)
   {
      ret = -1;
      goto end;
   }

   dec = linelet_decoder_new(width, height, &(const struct linelet_component) {
            .decomposition_levels = 7,
            .decimation = 0,
         }, 1,
         &(const struct linelet_decoder_options) {
            .precinct_samples = 256,
         });

   component = linelet_malloc(width * height * sizeof(int16_t));
   reference = linelet_malloc(width * height * sizeof(int16_t));
   out_component = linelet_malloc(width * height * sizeof(int16_t));
   encoded_buffer = linelet_malloc(width * height * 4);

   unsigned bits = 12;

   srand(1);
   for (unsigned c = 0; c < 4; c++)
   {
      switch (c & 3)
      {
         case 3:
            // Checkerboard pattern.
            for (unsigned i = 0; i < width * height; i++)
            {
               unsigned y = i / width;
               unsigned x = i % width;
               unsigned r = (x ^ y) & 1;
               component[i] = r ? ((1 << (bits - 1)) - 1) : -(1 << (bits - 1));
            }
            break;

         case 2:
            for (unsigned i = 0; i < width * height; i++)
            {
               unsigned x = i % width;
               unsigned r = x & 2;
               component[i] = r ? ((1 << (bits - 1)) - 1) : -(1 << (bits - 1));
            }
            break;

         case 1:
            // Random -/+ max range
            for (unsigned i = 0; i < width * height; i++)
            {
               unsigned r = rand() & 1;
               component[i] = r ? ((1 << (bits - 1)) - 1) : -(1 << (bits - 1));
            }
            break;

         case 0:
            // Random uniform
            for (unsigned i = 0; i < width * height; i++)
            {
               int16_t r = (int16_t)rand();
               component[i] = r >> (16 - bits);
            }
            break;
      }

      memcpy(reference, component, width * height * sizeof(int16_t));

      size_t target_size = linelet_encoder_process(enc,
            encoded_buffer, width * height * 4,
            (int16_t*[]) { component }, (const size_t[]) { width * sizeof(int16_t)});

      size_t decoded_size = linelet_decoder_process(dec, encoded_buffer,
            (int16_t*[]) { out_component }, (const size_t[]) { width * sizeof(int16_t) });

      if (target_size != decoded_size)
      {
         ret = -2;
         goto end;
      }

      if (memcmp(out_component, reference, width * height * sizeof(int16_t)))
      {
         ret = -3;
         goto end;
      }
   }

end:
   linelet_encoder_free(enc);
   linelet_decoder_free(dec);
   linelet_free(component);
   linelet_free(out_component);
   linelet_free(encoded_buffer);
   linelet_free(reference);
   return ret;
}

int encoder_test(void)
{
   int ret = 0;
   linelet_encoder_t *enc = NULL;
   linelet_decoder_t *dec = NULL;

   enc = linelet_encoder_new(64, 2,
         &(const struct linelet_component) {
            .decomposition_levels = 2,
            .decimation = 0,
         }, 1,
         &(const struct linelet_encoder_options) {
            .precinct_samples = 16,
         });

   if (!enc)
   {
      ret = -1;
      goto end;
   }

   dec = linelet_decoder_new(64, 2, &(const struct linelet_component) {
            .decomposition_levels = 2,
            .decimation = 0,
         }, 1,
         &(const struct linelet_decoder_options) {
            .precinct_samples = 16,
         });

   int16_t component[2][64] ALIGNED_STACK;
   int16_t out_component[2][64] ALIGNED_STACK = {{0}};
   for (unsigned i = 0; i < 64; i++)
   {
      component[0][i] = i;
      component[1][i] = i;
   }

   uint8_t encoded_buffer[sizeof(component)] ALIGNED_STACK;

   size_t target_size = linelet_encoder_process(enc,
         encoded_buffer, sizeof(encoded_buffer),
         (int16_t*[]) { &component[0][0] }, (const size_t[]) { 128 });

   size_t decoded_size = linelet_decoder_process(dec, encoded_buffer,
         (int16_t*[]) { &out_component[0][0] }, (const size_t[]) { 128 });

   if (target_size != decoded_size)
   {
      ret = -2;
      goto end;
   }

   for (unsigned i = 0; i < 64; i++)
   {
      component[0][i] = i;
      component[1][i] = i;
   }

   if (memcmp(out_component, component, sizeof(component)))
   {
      ret = -3;
      goto end;
   }

end:
   linelet_encoder_free(enc);
   linelet_decoder_free(dec);
   return ret;
}

